from copy import copy
from search_space.errors import ArgumentFunctionError, NotEvaluateError
from search_space.spaces.asts.naturals_values.__base__ import NaturalValuesNode
from search_space.utils import visitor
from search_space.spaces.asts import constraints
from . import VisitorLayer
from .visitor_natural_ast import NaturalAstVisitor


class DomainModifierVisitor(VisitorLayer):
    def __init__(self) -> None:
        super().__init__()
        self.natural_visitor = NaturalAstVisitor()

    @property
    def do_transform_to_check_sample(self):
        return False

    def domain_optimization(self, node, domain):
        self.domain, self._context = domain, None
        return self.visit(node)

    def transform_to_modifier(self, node, domain, context):
        self.domain, self._context = domain, context
        return self.visit(node)

    def __op_apply__(self, a, b, op):
        if a is None:
            self.domain = op(self.domain, b)

        elif b is None:
            self.domain = op(a, self.domain)

        else:
            return op(b, a)

    @visitor.on("node")
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node: constraints.AstRoot):

        for n in node.asts:
            self.visit(n)

        return node, self.domain

    #################################################################
    #                                                               #
    #                 Arithmetic Visit                              #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AddOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x + y)

    @visitor.when(constraints.SubOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x - y)

    @visitor.when(constraints.MultiOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x * y)

    @visitor.when(constraints.DivOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x / y)

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AndOp)
    def visit(self, node):
        a = self.visit(node.target)

        if type(a) == bool and not a:
            return

        b = self.visit(node.other)

        if not None in [a, b]:
            return target and limit

    @visitor.when(constraints.OrOp)
    def visit(self, node):
        current_domain = copy(self.domain)
        a = self.visit(node.target)

        if type(a) == bool and a:
            return

        current_domain, self.domain = self.domain, current_domain
        b = self.visit(node.other)

        if type(a) == bool and b is None:
            return

        if not None in [a, b]:
            return target or limit

        self.domain = self.domain | current_domain

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################
    @visitor.when(constraints.LessOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x < y)

    @visitor.when(constraints.LessOrEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x <= y)

    @visitor.when(constraints.GreatOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x > y)

    @visitor.when(constraints.GreatOrEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x >= y)

    @visitor.when(constraints.NotEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x != y)

    @visitor.when(constraints.EqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x == y)

    #################################################################
    #                                                               #
    #                 Segmentation Visit                            #
    #                                                               #
    #################################################################

    @visitor.when(constraints.SegmentationEqualOp)
    def visit(self, node: constraints.SegmentationEqualOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor == value

        self.domain = self.domain.__mod_eq__(factor, value)

    @visitor.when(constraints.SegmentationNotEqualOp)
    def visit(self, node: constraints.SegmentationNotEqualOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor != value

        self.domain = self.domain.__mod_neq__(factor, value)

    @visitor.when(constraints.SegmentationLessOp)
    def visit(self, node: constraints.SegmentationLessOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor < value

        self.domain = self.domain.__mod_lt__(factor, value)

    @visitor.when(constraints.SegmentationLessOrEqualOp)
    def visit(self, node: constraints.SegmentationLessOrEqualOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor <= value

        self.domain = self.domain.__mod_le__(factor, value)

    @visitor.when(constraints.SegmentationGreatOp)
    def visit(self, node: constraints.SegmentationGreatOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor > value

        self.domain = self.domain.__mod_gt__(factor, value)

    @visitor.when(constraints.SegmentationGreatOrEqualOp)
    def visit(self, node: constraints.SegmentationGreatOrEqualOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)
        value = self.visit(node.value)

        if not target is None:
            return target % factor >= value

        self.domain = self.domain.__mod_ge__(factor, value)

    @visitor.when(constraints.SegmentationAddOp)
    def visit(self, node: constraints.SegmentationAddOp):
        target = self.visit(node.target)
        factor = self.visit(node.value)

        if not target is None:
            return target + factor

        self.domain = self.domain.__mod_add__(factor, 0)

    @visitor.when(constraints.SegmentationSubOp)
    def visit(self, node: constraints.SegmentationSubOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)

        if not target is None:
            return target - factor

        self.domain = self.domain.__mod_sub__(factor, 0)

    @visitor.when(constraints.SegmentationMultiOp)
    def visit(self, node: constraints.SegmentationMultiOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)

        if not target is None:
            return target * factor

        self.domain = self.domain.__mod_mult__(factor, 0)

    @visitor.when(constraints.SegmentationDivOp)
    def visit(self, node: constraints.SegmentationDivOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)

        if not target is None:
            return target / factor

        self.domain = self.domain.__mod_div__(factor, 0)

    @visitor.when(constraints.SegmentationModOp)
    def visit(self, node: constraints.SegmentationModOp):
        target = self.visit(node.target)
        factor = self.visit(node.other)

        if not target is None:
            return target % factor

        self.domain = self.domain.__mod_eq__(factor, 0)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    # @visitor.when(constraints.GetAttr)
    # def visit(self, node: constraints.GetAttr):
    #     target = self.visit(node.target)
    #     other = self.visit(node.other)

    #     return target.__dict__[other]

    @visitor.when(constraints.GetItem)
    def visit(self, node: constraints.GetItem):
        target = self.visit(node.target)
        other = self.visit(node.other)

        return target[other]

    @visitor.when(constraints.SelfNode)
    def visit(self, node):
        pass

    @visitor.when(constraints.NaturalValue)
    def visit(self, node):
        if isinstance(node.target, NaturalValuesNode):
            return self.natural_visitor.get_value(
                node.target, context=self._context
            )

        return node.target

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg)

        return node.func(*new_args, **new_kw)

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.FunctionNode):
        raise ArgumentFunctionError()

    @visitor.when(constraints.NotEvaluate)
    def visit(self, node):
        raise NotEvaluateError()



from typing import Iterator
from search_space.utils import visitor
from search_space.spaces.asts import constraints
from search_space.spaces.asts import naturals_values
from search_space.errors import InvalidSampler, NotEvaluateError
from search_space.utils.ast_tools import index_list
from . import VisitorLayer
from search_space.utils.singleton import Singleton
from .visitor_natural_ast import NaturalAstVisitor, NaturalValuesNode
from search_space.utils import itertools
from typing import Iterable
from search_space.utils.itertools import is_iterable


class ValidateSampler(VisitorLayer, metaclass=Singleton):

    def __init__(self) -> None:
        self.natural_visitor = NaturalAstVisitor()

    def transform_to_check_sample(self, node, sample, context):
        self.sample, self.context = sample, context
        return self.visit(node)

    @visitor.on("node")
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node, current_index=[]):
        for n in node.asts:
            try:
                result = self.visit(n, current_index)
                if not result:
                    raise InvalidSampler("False Restriction")
            except NotEvaluateError:
                pass

    #################################################################
    #                                                               #
    #                 Arithmetic Visit                              #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AddOp)
    def visit(self, node, current_index=[]):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a + b

    @visitor.when(constraints.SubOp)
    def visit(self, node, current_index=[]):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a - b

    @visitor.when(constraints.MultiOp)
    def visit(self, node, current_index=[]):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a * b

    @visitor.when(constraints.DivOp)
    def visit(self, node, current_index=[]):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a / b

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AndOp)
    def visit(self, node, current_index=[]):
        a = self.visit(node.target, current_index)

        if type(a) == bool and not a:
            return a

        return self.visit(node.other, current_index)

    @visitor.when(constraints.OrOp)
    def visit(self, node, current_index=[]):
        a = False
        try:
            a = self.visit(node.target, current_index)
        except InvalidSampler:
            pass

        if a:
            return a

        return self.visit(node.other, current_index)

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.NotEqualOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b) and not is_iterable(a):
            if not a in b:
                return self

        elif is_iterable(a) and not is_iterable(b):
            if not b in a:
                return self

        elif a != b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} != {b} ]")

    @visitor.when(constraints.EqualOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b) and not is_iterable(a):
            if a in b:
                return self

        elif is_iterable(a) and not is_iterable(b):
            if b in a:
                return self

        elif a == b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} == {b} ]")

    @visitor.when(constraints.LessOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = min(b)

        if is_iterable(a):
            a = max(a)

        if a < b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} < {b} ]")

    @visitor.when(constraints.LessOrEqualOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = min(b)

        if is_iterable(a):
            a = max(a)

        if a <= b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} < {b} ]")

    @visitor.when(constraints.GreatOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = max(b)

        if is_iterable(a):
            a = min(a)

        if a > b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} > {b} ]")

    @visitor.when(constraints.GreatOrEqualOp)
    def visit(self, node, current_index=[]):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = max(b)

        if is_iterable(a):
            a = min(a)

        if a >= b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} >= {b} ]")

    #################################################################
    #                                                               #
    #                 Segmentation Visit                            #
    #                                                               #
    #################################################################

    @visitor.when(constraints.SegmentationEqualOp)
    def visit(self, node: constraints.SegmentationEqualOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        mod = abs(target % factor)
        if is_iterable(value) and mod in value:
            return self

        if mod == abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} == {value} ]")

    @visitor.when(constraints.SegmentationNotEqualOp)
    def visit(self, node: constraints.SegmentationNotEqualOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        mod = target % factor
        if (
            is_iterable(value)
            and not mod in value
            and not -mod in value
        ):
            return self

        if mod != value and -mod != value:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} != {value} ]")

    @visitor.when(constraints.SegmentationLessOp)
    def visit(self, node: constraints.SegmentationLessOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = min(value)

        if abs(target % factor) < abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} < {value} ]")

    @visitor.when(constraints.SegmentationLessOrEqualOp)
    def visit(self, node: constraints.SegmentationLessOrEqualOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = min(value)

        if abs(target % factor) <= abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} <= {value} ]")

    @visitor.when(constraints.SegmentationGreatOp)
    def visit(self, node: constraints.SegmentationGreatOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = max(value)

        if abs(target % factor) > abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} > {value} ]")

    @visitor.when(constraints.SegmentationGreatOrEqualOp)
    def visit(self, node: constraints.SegmentationGreatOrEqualOp, current_index=[]):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = max(value)

        if abs(target % factor) >= abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} >= {value} ]")

    # @visitor.when(constraints.SegmentationAddOp)
    # def visit(self, node: constraints.SegmentationAddOp, current_index=[]):
    #     target = self.visit(node.target, current_index)
    #     factor = self.visit(node.value, current_index)

    #     if not target is None:
    #         return target + factor

    #     self.domain = self.domain.__mod_add__(factor, 0)

    # @visitor.when(constraints.SegmentationSubOp)
    # def visit(self, node: constraints.SegmentationSubOp, current_index=[]):
    #     target = self.visit(node.target, current_index)
    #     factor = self.visit(node.other, current_index)

    #     if not target is None:
    #         return target - factor

    #     self.domain = self.domain.__mod_sub__(factor, 0)

    # @visitor.when(constraints.SegmentationMultiOp)
    # def visit(self, node: constraints.SegmentationMultiOp, current_index=[]):
    #     target = self.visit(node.target, current_index)
    #     factor = self.visit(node.other, current_index)

    #     if not target is None:
    #         return target * factor

    #     self.domain = self.domain.__mod_mult__(factor, 0)

    # @visitor.when(constraints.SegmentationDivOp)
    # def visit(self, node: constraints.SegmentationDivOp, current_index=[]):
    #     target = self.visit(node.target, current_index)
    #     factor = self.visit(node.other, current_index)

    #     if not target is None:
    #         return target / factor

    #     self.domain = self.domain.__mod_div__(factor, 0)

    # @visitor.when(constraints.SegmentationModOp)
    # def visit(self, node: constraints.SegmentationModOp, current_index=[]):
    #     target = self.visit(node.target, current_index)
    #     factor = self.visit(node.other, current_index)

    #     if not target is None:
    #         return target % factor

    #     self.domain = self.domain.__mod_eq__(factor, 0)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.GetItem)
    def visit(self, node: constraints.GetItem):
        target = self.visit(node.target)

        try:
            _len = len(target)
        except TypeError:
            raise InvalidSampler(
                f"inconsistent sampler => {target} isn't indexed")

        indexes = self.visit_natural(node.other, _len)

        def f(a, b):
            try:
                return a[b]
            except TypeError:
                raise InvalidSampler(
                    f"inconsistent sampler => sample isn't indexed")

        return itertools.map_reduce(target, indexes, f)

    @visitor.when(constraints.GetAttr)
    def visit(self, node: constraints.GetAttr):
        b = self.visit(node.other)
        a = self.visit(node.target)

        def f(a, b):
            try:
                space = getattr(a, b)
            except AttributeError:
                raise InvalidSampler(
                    f"inconsistent sampler => {a} don't has {b} member")

            try:
                sampler = space.get_sample
            except AttributeError:
                raise InvalidSampler(
                    f"inconsistent sampler => {a}.{b} isn't search space")

            return sampler(context=a.__instance_context__)[0]

        return itertools.map_reduce(a, b, f)

    @ visitor.when(constraints.SelfNode)
    def visit(self, node):
        return self.sample

    @ visitor.when(constraints.NaturalValue)
    def visit(self, node):
        if isinstance(node.target, NaturalValuesNode):
            return self.natural_visitor.get_value(
                node.target,
                context=self.context,
            )
        return node.target

    # @visitor.when(constraints.FunctionNode)
    # def visit(self, node: constraints.FunctionNode, current_index):
    #     new_args = []
    #     for arg in node.args:
    #         new_args.append(self.visit(arg, current_index))

    #     new_kw = {}
    #     for name, arg in node.kwargs:
    #         new_kw[name] = self.visit(arg, current_index)

    #     result = node.func(*new_args, **new_kw)
    #     return result

    # @visitor.when(constraints.AdvancedFunctionNode)
    # def visit(self, node: constraints.FunctionNode, current_index):
    #     new_args = []
    #     for arg in node.args:
    #         new_args.append(self.visit(arg, current_index))

    #     new_kw = {}
    #     for name, arg in node.kwargs:
    #         new_kw[name] = self.visit(arg, current_index)

    #     result = node.func(*new_args, **new_kw)
    #     return result

    #################################################################
    #                                                               #
    #                 Natural Visit                                 #
    #                                                               #
    #################################################################

    @visitor.on("node")
    def visit_natural(self, node, _len=None):
        pass

    @visitor.when(constraints.NaturalValue)
    def visit_natural(self, node, _len):
        result = node.target
        if isinstance(node.target, NaturalValuesNode):
            self.len = _len
            result = self.visit_natural(node.target)
            self.len = None

        return result

    #################################################################
    #                                                               #
    #                 Arithmetic Visit                              #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.ModOp)
    def visit(self, node: naturals_values.ModOp):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce_longest(a, b, lambda a, b: a % b, 1)

    @visitor.when(naturals_values.AddOp)
    def visit(self, node: naturals_values.AddOp):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce_longest(a, b, lambda a, b: a + b, 0)

    @visitor.when(naturals_values.SubOp)
    def visit(self, node: naturals_values.SubOp):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce_longest(a, b, lambda a, b: a - b, 0)

    @visitor.when(naturals_values.MultiOp)
    def visit(self, node: naturals_values.MultiOp):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce_longest(a, b, lambda a, b: a * b, 1)

    @visitor.when(naturals_values.DivOp)
    def visit(self, node: naturals_values.DivOp):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce_longest(a, b, lambda a, b: a / b, 1)

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.AndOp)
    def visit(self, node: naturals_values.AndOp):
        a = self.visit_natural(node.a)

        if not isinstance(a, Iterator) and not a:
            return True

        b = self.visit_natural(node.b)

        return itertools.compare(a, b, lambda a, b: a and b)

    @visitor.when(naturals_values.OrOp)
    def visit(self, node: naturals_values.OrOp):
        a = self.visit_natural(node.a)

        if not isinstance(a, Iterator) and a:
            return True

        b = self.visit_natural(node.b)

        return itertools.compare(a, b, lambda a, b: a or b)

    #################################################################
    #                                                               #
    #             Binary Cmp Natural Visit                          #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.LessOp)
    def visit_natural(self, node: naturals_values, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a < b)

    @visitor.when(naturals_values.LessOrEqualOp)
    def visit_natural(self, node: naturals_values.LessOrEqualOp, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a <= b)

    @visitor.when(naturals_values.GreatOp)
    def visit_natural(self, node: naturals_values.GreatOp, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a > b)

    @visitor.when(naturals_values.GreatOrEqualOp)
    def visit_natural(self, node: naturals_values.GreatOrEqualOp, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a >= b)

    @visitor.when(naturals_values.NotEqualOp)
    def visit_natural(self, node: naturals_values.NotEqualOp, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a != b)

    @visitor.when(naturals_values.EqualOp)
    def visit_natural(self, node: naturals_values.EqualOp, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.filter(a, b, lambda a, b: a == b)

    #################################################################
    #                                                               #
    #                 Simple Natural Visit                          #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.GetAttr)
    def visit_natural(self, node: naturals_values.GetAttr, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        def f(a, b):
            value = getattr(a, b)
            try:
                _ = value.get_sample
            except AttributeError:
                return value

            return value.get_sample(a.__instance_context__)[0]

        yield itertools.map_reduce(a, b, f)

    @visitor.when(naturals_values.GetItem)
    def visit_natural(self, node: naturals_values.GetItem, _len=None):
        a = self.visit_natural(node.a)
        b = self.visit_natural(node.b)

        yield itertools.map_reduce(a, b, lambda a, b: a[b])

    @visitor.when(naturals_values.IndexSelf)
    def visit_natural(self, node, _len=None):
        for i in range(self.len - 1):
            yield i

        return self.len

    @visitor.when(naturals_values.SpaceSelfNode)
    def visit_natural(self, node: naturals_values.SpaceSelfNode, _len=None):
        return node.a.get_sample(self.context)[0]

    @visitor.when(naturals_values.NaturalValue)
    def visit_natural(self, node: naturals_values.NaturalValue, _len=None):
        return node.a



from xml import dom
from search_space.errors import CircularDependencyDetected, DetectedRuntimeDependency, NotEvaluateError
from ..asts import constraints
from search_space.utils import visitor
from search_space.spaces.asts import constraints, naturals_values
from . import VisitorLayer
from .visitor_natural_ast import NaturalAstVisitor, NaturalValuesNode


class MemberAstModifierVisitor(VisitorLayer):
    def __init__(self, space, member) -> None:
        self.member = member
        self.space = space
        self.natural_visitor = NaturalAstVisitor()

    @property
    def context(self):
        try:
            if self._context != None:
                return self._context
        except AttributeError:
            pass

        raise DetectedRuntimeDependency()

    def domain_optimization(self, node, domain):
        self._context = None
        return self.visit(node), domain

    def transform_to_modifier(self, node, domain=None, context=None):
        self._context = context
        return self.visit(node), domain

    def transform_to_check_sample(self, node, sample, context=None):
        self._context = context
        return self.visit(node, context=context)

    @visitor.on("node")
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node):
        result = constraints.AstRoot([])

        for n in node.asts:
            result.add_constraint(
                self.visit(n))

        return result

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.UniversalVariableBinaryOperation)
    def visit(self, node):
        a = self.visit(node.target)
        b = self.visit(node.other)

        if isinstance(node, constraints.SegmentationExprNode):
            c = self.visit(node.value)
            return type.__call__(node.__class__, a, b, c)

        return type.__call__(node.__class__, a, b)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.GetItem)
    def visit(self, node):
        index = self.visit(node.other)
        target = self.visit(node.target)

        if type(index.target) == bool:
            return target

        return constraints.GetItem(target, index)

    @visitor.when(constraints.GetAttr)
    def visit(self, node: constraints.GetAttr):
        target = self.visit(node.target)
        other = self.visit(node.other)

        if target.is_self and isinstance(other, constraints.NaturalValue):
            if other.target == self.member:
                return target
            else:
                return constraints.NaturalValue(naturals_values.SpaceSelfNode(self.space[other.target]))

        return constraints.GetAttr(target, other)

    @visitor.when(constraints.SelfNode)
    def visit(self, node):
        return node

    @visitor.when(constraints.NaturalValue)
    def visit(self, node):
        if isinstance(node.target, NaturalValuesNode):
            try:
                result = self.natural_visitor.get_value(
                    node.target, context=self._context
                )
                return constraints.NaturalValue(result)
            except CircularDependencyDetected:
                return constraints.NotEvaluate()

        return node

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg)

        return constraints.FunctionNode(node.func, new_args, new_kw)

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.AdvancedFunctionNode):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg)

        return constraints.AdvancedFunctionNode(
            (node.args_target, node.kw_target),
            node.func, new_args, new_kw,
            node.cls
        )



from .__base__ import NaturalValuesNode

from .logic_ops import AndOp, OrOp
from .expr_ops import EqualOp, NotEqualOp, LessOp, LessOrEqualOp, GreatOp, GreatOrEqualOp
from .expr_ops import AddOp, SubOp, ModOp, DivOp, MultiOp
from .expr_ops import IndexSelf
from .atoms import GetAttr, GetItem
from .atoms import NaturalValue, SpaceSelfNode

"""
    NaturalValues := Cmp | Constraint
                  := Cmp & Constraint
                  := Cmp

             Expr := Expr == Expr 
             Expr := Expr != Expr 
             Expr := Expr >= Expr 
             Expr := Expr >  Expr 
             Expr := Expr <= Expr 
             Expr := Expr <  Expr 
          
             Expr := Expr + Expr
             Expr := Expr - Expr
             Expr := Expr * Expr
             Expr := Expr / Expr
             Expr := Expr % Expr
             
             Expr := Atom
             Expr := SelfIndex

             Atom := Atom[index]
             Atom := Atom[member]
             Atom := NaturalValue
             Atom := (Constraint)
             Atom := SpaceExpr
"""



from search_space.spaces import SearchSpace, BasicSearchSpace
from search_space.sampler.distribution_names import UNIFORM
from search_space.spaces.visitors import visitors
from search_space.spaces.domains.continuos_domain import ContinuosDomain
from search_space.spaces.domains.natural_domain import NaturalDomain
from .space_manager import SpacesManager
from search_space.utils.infinity import oo

__all__ = [
    "BasicContinueSearchSpace",
    "BasicNaturalSearchSpace",
]


@SpacesManager.registry(float)
class BasicContinueSearchSpace(BasicSearchSpace):
    def __init__(self, min=-oo, max=oo, distribute_like=UNIFORM) -> None:
        super().__init__(ContinuosDomain(min, max), distribute_like)


@SpacesManager.registry(int)
class BasicNaturalSearchSpace(BasicSearchSpace):
    def __init__(self, min=-oo, max=oo, distribute_like=UNIFORM) -> None:
        super().__init__(NaturalDomain(min, max), distribute_like)



from copy import copy
import inspect
from typing_extensions import Self
from search_space.context_manager.sampler_context import SamplerContext
from search_space.sampler.distribution_names import UNIFORM
from search_space.spaces.visitors import visitors
from search_space.spaces import BasicSearchSpace
import imp
from typing import List, Type
from search_space.utils.singleton import Singleton
from search_space.spaces import SearchSpace
from search_space.spaces.asts import constraints as ast_constraint
from typing import _UnionGenericAlias
from search_space.spaces.domains.categorical_domain import CategoricalDomain


class SpacesManager(metaclass=Singleton):
    def __init__(self) -> None:
        self.__spaces = {}
        self.__metadata = {}

    @staticmethod
    def registry(_type: Type = None):
        def f(cls):
            manager = SpacesManager()
            if _type is None:
                manager.__spaces[cls] = cls
            else:
                manager.__spaces[_type] = cls

            return cls
        return f

    def get_space_by_type(self, _type: Type) -> Type[SearchSpace]:
        try:
            return self.__spaces[_type]
        except KeyError:
            if type(_type) == _UnionGenericAlias:
                args = [self.get_space_by_type(t) for t in _type.__args__]
                return UnionSpace(*args)

            def f(*args, **kwargs):

                if _type == type(None):
                    return NoneSpace()

                if _type == Self:
                    return SelfSpace()

                return SpaceFactory(_type)
            return f

    def find_metadata(self, _type):
        return self.__metadata[_type]

    def save_metadata(self, _type, sub_space):
        self.__metadata[_type] = {}
        for name, method in inspect.getmembers(_type, inspect.isfunction):
            self.__metadata[_type][name] = ClassFunction(
                method, sub_space, _type
            )

        return self.__metadata[_type]


class FunctionParamInfo:
    def __init__(self, name) -> None:
        self.name = name
        self.type = None
        self.default = None

    def type_init(self):
        if self.default is None and not self.type is None:
            self.default = SpacesManager().get_space_by_type(self.type)()

    def __eq__(self, __o: object) -> bool:
        return __o == self.name

    def get_sample(self, context=None, sub_space=[]):
        if self.default is None:
            return None

        try:
            _ = self.default.get_sample
        except AttributeError:
            return self.default

        try:
            value = [s for s in sub_space if hash(
                s) == hash(self.default)][0]
            name = value.space_name
            value.space_name = self.default.space_name
        except IndexError:
            name = self.default.space_name
            value = self.default

        result, _ = value.get_sample(context=context)
        value.space_name = name
        return result


class ClassFunction:
    def __init__(self, func, sub_space: list, _self, context=None, params=None, instance=None) -> None:
        self.func = func
        self.params = params
        self.context = context
        self.sub_space = []
        self.instance = instance

        if params is None:
            func_data = inspect.getfullargspec(func)
            self.params: List[FunctionParamInfo] = []

            for arg in func_data.args:
                if arg == 'self':
                    continue
                self.params.append(FunctionParamInfo(arg))

            for param in self.params:
                try:
                    param.type = func_data.annotations[param.name]
                except KeyError:
                    pass

            if not func_data.defaults is None:
                for i, value in enumerate(reversed(func_data.defaults)):

                    try:
                        value = value.space
                    except AttributeError:
                        pass

                    _value = copy(value)

                    try:
                        member = [s for s in sub_space if hash(
                            s) == hash(value)][0]
                        _value.set_hash(hash(member))
                    except IndexError:
                        pass
                    finally:
                        value = _value
                    try:
                        value.__self_assign__(_self)
                    except AttributeError:
                        pass

                    try:
                        if not value.space_name.startswith(f'{self.params[-1 -i].name}:'):
                            value.space_name = f'{self.params[-1 -i].name}:{value.space_name}'
                    except:
                        pass

                    self.params[-1 - i].default = value

            for arg in self.params:
                arg.type_init()

    def get_new_func(self, context, sub_space, instance=None):
        result = ClassFunction(self.func, [], None,
                               context, self.params, instance)
        result.sub_space = sub_space

        return result

    def sample_params(self, *args, **kwds):
        new_args = []
        args_index = 0

        for arg in self.params:
            try:
                new_args.append(kwds[arg.name])
            except KeyError:
                if args_index < len(args):
                    new_args.append(args[args_index])
                    args_index += 1
                else:
                    new_args.append(arg.get_sample(
                        context=self.context,
                        sub_space=self.sub_space)
                    )
        return new_args

    def __call__(self, *args, **kwds):
        inner_context = self.context
        self.context = self.context.create_child(
            f'called_{self.func.__name__}')

        try:
            new_args = self.sample_params(*args, **kwds)
        except Exception as e:
            error = e

        self.context = inner_context

        try:
            raise error
        except NameError:

            return self.func(*([self.instance] + new_args))


class SpaceFactory(SearchSpace):
    def __init__(
        self, _type,
        distribute_like=None,
        sampler=None,
        ast=None,
        clean_asts=None,
        layers=[]
    ) -> None:

        super().__init__(
            _type,
            distribute_like,
            sampler,
            ast_constraint.AstRoot() if ast is None else ast,
            ast_constraint.AstRoot() if ast is None else clean_asts,
            layers
        )

        self.type = _type
        self._sub_space = {}
        self.space_name = _type.__name__

        for name, member in inspect.getmembers(self.type):
            try:
                if isinstance(member.space, BasicSearchSpace):
                    self._sub_space[name] = copy(member.space)
                    self._sub_space[name].set_hash(hash(member.space))
                    self._sub_space[name].layers_append(
                        visitors.EvalAstChecked(),
                        visitors.MemberAstModifierVisitor(self, name)
                    )

            except AttributeError:
                pass

        manager = SpacesManager()
        try:
            self._metadata = manager.find_metadata(self.type)
        except KeyError:
            self._metadata = manager.save_metadata(
                self.type, list(self._sub_space.values()))

        self.visitor_layers = []

    def __ast_init_filter__(self):
        return self.ast_constraint + self._clean_asts

    def __ast_result_filter__(self, result):
        return result

    def __sampler__(self, domain, context: SamplerContext):
        instance_context = context.create_child(f'{self.space_name}_members')
        key_list = list(self._sub_space.values())
        class_func = self._metadata['__init__'].get_new_func(
            instance_context, key_list)

        class_instance = self.type(*class_func.sample_params())
        class_instance.__instance_context__ = instance_context

        for name, class_func in self._metadata.items():
            if name == '__init__':
                continue

            setattr(class_instance, name,
                    class_func.get_new_func(instance_context, key_list, class_instance))

        return class_instance, instance_context

    def __ast_optimization__(self, ast_list):

        super().__ast_optimization__(ast_list)

        for key, space in self._sub_space.items():
            self._sub_space[key] = space.__ast_optimization__(ast_list)

        return self

    def __save_ast_from_self__(self, ast_list):
        super().__ast_optimization__(ast_list)
        return self

    def __getitem__(self, index):
        return self._sub_space[index]

    def layers_append(self, *args):
        super().layers_append(*args)

        for space in self._sub_space.values():
            space.layers_append(*args)

    def __repr__(self) -> str:
        return f'Space({self.space_name})'


class SelfSpace:
    def __init__(self, distribute_like=None) -> None:
        self.self_space = None
        self.ast_cache = []

    def get_sample(self, context=None, local_domain=None):
        return self.self_space.get_sample()[0], context

    def __getattr__(self, __name: str):
        return getattr(self.self_space, __name)

    def __self_assign__(self, space):
        if self.self_space is None:
            self.self_space = SpacesManager().get_space_by_type(space)()
            for ast in self.ast_cache:
                self.self_space = self.self_space.__ast_optimization__(ast)

    def __ast_optimization__(self, ast_list):
        if self.self_space is None:
            self.ast_cache.append(ast_list)
        else:
            self.self_space = self.self_space.__save_ast_from_self__(ast_list)

        return self


class NoneSpace(BasicSearchSpace):
    def __init__(self) -> None:
        super().__init__(None, None)

    def get_sample(self, context=None, local_domain=None):
        return None, context

    def __ast_optimization__(self, ast_list):
        return self

    def __eq__(self, __o: object) -> bool:
        return __o == None or super().__eq__(__o)


class UnionSpace(SearchSpace):
    def __init__(
        self,
        *domain,
        distribute_like=UNIFORM,
        sampler=None,
        ast=None,
        clean_asts=None,
        layers=[]
    ) -> None:

        super().__init__(
            CategoricalDomain(domain),
            distribute_like,
            sampler,
            ast_constraint.AstRoot() if ast is None else ast,
            ast_constraint.AstRoot() if ast is None else clean_asts,
            layers=[visitors.TypeDomainModifierVisitor()] if len(
                layers) == 0 else layers
        )

    def get_sample(self, context=None, local_domain=None):
        space, context = super().get_sample(context)
        return space.get_sample(context)

        # try:
        # except RecursionError:
        #     if None in self.initial_domain.limits:
        #         warn('ForceResult: For RecursionError Select None Value')
        #         context.registry_sampler(space, None)
        #         return None, context
        #     else:
        #         raise RecursionErrorSpaceDefinition

    def __call__(self, *args, **kwds):
        self.initial_domain = CategoricalDomain(
            [space(*args, **kwds) for space in self.initial_domain.list]
        )

        return self

    def __ast_optimization__(self, ast_list):

        super().__ast_optimization__(ast_list)

        result = []
        for space in self.initial_domain.list:
            result.append(space.__ast_optimization__(ast_list))

        self.initial_domain.list = result

        return self

    def __self_assign__(self, space):

        for s in self.initial_domain.list:
            try:
                s.__self_assign__(space)
            except AttributeError:
                pass

    def __check_sample__(self, sample, ast_result, context):
        return

    def __copy__(self):

        result = type(self)(
            distribute_like=self.__distribute_like__,
            sampler=None,
            ast=ast_constraint.AstRoot(copy(self.ast_constraint.asts)),
            clean_asts=ast_constraint.AstRoot(copy(self._clean_asts.asts)),
            layers=copy(self.visitor_layers)
        )

        result.initial_domain = copy(self.initial_domain)
        return result



from ..asts import constraints


class VisitorLayer:

    @property
    def do_domain_optimization(self):
        return True

    @property
    def do_transform_to_modifier(self):
        return True

    @property
    def do_transform_to_check_sample(self):
        return True

    def domain_optimization(self, node, domain):
        return node, domain

    def transform_to_modifier(self, node, domain, context):
        return node, domain

    def transform_to_check_sample(self, node, sample, context):
        return node


# TODO: write visitor



from .expr_ops import ExprOpsNode
from .__base__ import check_natural_value, UniversalVariableNode
from search_space.utils.ast_tools import decorated_all_methods
from search_space.errors import UnSupportOpError
from . import __namespace__ as nsp


def natural_node_check(fn):
    def f(*args):
        if not isinstance(args[1], NaturalValue):
            raise UnSupportOpError(args[0], args[1], 'members')

        return fn(*args)
    f.__name__ = fn.__name__
    return f


@decorated_all_methods(check_natural_value)
@decorated_all_methods(natural_node_check)
class AtomsOp(ExprOpsNode):

    def __getattr__(self, name):
        return GetAttr(self, name)

    def __getitem__(self, index):
        return GetItem(self, index)


class GetAttr(AtomsOp):
    pass


class GetItem(AtomsOp):
    pass


class Atoms(AtomsOp):
    def __init__(self, target) -> None:
        super().__init__(target, None)


class NaturalValue(Atoms, metaclass=nsp.NaturalValueNode):
    pass


class SelfNode(Atoms):
    def __init__(self) -> None:
        super().__init__(None)

    @ property
    def is_self(self):
        return True


class NotEvaluate(Atoms):
    def __init__(self) -> None:
        super().__init__(None)


class FunctionNode(NaturalValue):
    def __init__(self, func, args, kwargs) -> None:
        super().__init__(None)
        self.func = func
        self.args = args
        self.kwargs = kwargs


class AdvancedFunctionNode(FunctionNode):
    def __init__(self, target, func, args, kwargs, cls=None) -> None:
        super().__init__(func, args, kwargs)
        self.cls = cls
        self.args_target, self.kw_target = target


def params_analyze(func, args, kwargs, cls=None):
    new_args = []
    new_kwargs = {}
    var_intro_arg = []
    var_intro_kw = []

    for i, arg in enumerate(args):
        if isinstance(arg, UniversalVariableNode):
            var_intro_arg.append(i)
            new_args.append(arg)
        else:
            new_args.append(NaturalValue(arg))

    for name, arg in kwargs.items():
        if isinstance(arg, UniversalVariableNode):
            var_intro_kw.append(name)
            new_kwargs[name] = args
        else:
            new_kwargs[name] = NaturalValue(arg)

    if len(var_intro_arg) + len(var_intro_kw) > 0:
        return AdvancedFunctionNode((var_intro_arg, var_intro_kw), func, new_args, kwargs, cls)

    return FunctionNode(func, new_args, new_kwargs)


class AdvancedSelectorMetaClass(type):
    def __call__(self, *args, **kwds):
        instance = super().__call__(*args, **kwds)
        return params_analyze(instance.__call__, args, kwds, cls=instance)


class FunctionalConstraintClass(metaclass=AdvancedSelectorMetaClass):
    def __init__(self, *args, **kwgs) -> None:
        pass


class FunctionalConstraintFactory:
    def __mro_entries__(self, x):
        return FunctionalConstraintClass, object,

    def __call__(self, func):
        def f(*args, **kwargs):
            return params_analyze(func, args, kwargs)
        return f


FunctionalConstraint = FunctionalConstraintFactory()



from search_space.spaces import FunctionalConstraint


__all__ = [
    "Factorial"
]


@FunctionalConstraint
def Factorial(x: int):
    result = 1
    for i in range(2, x + 1):
        result *= i

    return result



from search_space.errors import InvalidSpaceConstraint, InvalidSpaceDefinition
from search_space.spaces.domains.categorical_domain import CategoricalDomain
from search_space.spaces.domains.__base__ import NumeralDomain
from search_space.spaces.domains.linear_transform_domain import LinearTransformedDomain
from .bached_domain import BachedDomain, LogBachedDomain
from search_space.sampler import Sampler
from . import __namespace__ as nsp
from typing import Iterable
from search_space.utils.itertools import is_iterable


class NaturalDomain(NumeralDomain, metaclass=nsp.NaturalDomain):
    def __init__(self, _min, _max) -> None:
        self.min, self.max = _min, _max

    def __copy__(self):
        return NaturalDomain(self.min, self.max)

    def get_sample(self, sampler: Sampler):
        return sampler.get_int(self.min, self.max)

    @property
    def limits(self):
        return (self.min, self.max)

    #################################################################
    #                                                               #
    #                     List Operations                           #
    #                                                               #
    #################################################################

    def __getitem__(self, index):
        raise InvalidSpaceConstraint("Numeral value isn't indexed")

    #################################################################
    #                                                               #
    #                 Binary Compare Operations                     #
    #                                                               #
    #################################################################

    def __eq__(self, other):
        if is_iterable(other):
            return CategoricalDomain([item for item in other if self.min <= item and item <= self.max])

        if other > self.max or other < self.min:
            raise InvalidSpaceDefinition(
                f"{other} isn't intro the [{self.min}, {self.max}]")

        self.max = other
        self.min = other

        return self

    # def __ne__(self, other):
    #     if is_iterable(other):
    #         if len(other) == 0:
    #             return self
    #         other = [item for item in other if self.min <=
    #                  item and item <= self.max]

    #         other.sort()
    #         try:
    #             other = [other[0]] + [other[i] for i in range(1, len(other))
    #                                   if other[i] - other[i-1] > 0.51]

    #             while len(other) > 0 and self.min == other[0]:
    #                 other.pop(0)
    #                 self.min += 0.51

    #             while len(other) > 0 and self.max == other[-1]:
    #                 other.pop(-1)
    #                 self.max -= 0.51
    #         except IndexError:
    #             raise InvalidSpaceDefinition('NaturalDomain is empty')

    #         other = [self.min - 0.51] + other + [self.max + 0.51]
    #         return BachedDomain(
    #             * [NaturalDomain(other[i-1] + 0.51, other[i] - 0.51) for i in range(1, len(other))]
    #         )

    #     "If other is out of current domain, it will never enter and current domain is right"
    #     if other > self.max or other < self.min:
    #         return self

    #     return BachedDomain(NaturalDomain(self.min, other - 0.51), NaturalDomain(other + 0.51, self.max))

    def __ne__(self, other):
        """
        This operation don't have to computing in less than nlogn.
        Is the only way that we can create the correctly baches 
        """

        if not is_iterable(other):
            other = [other]

        other = [item for item in other
                 if self.min <= item and item <= self.max]
        other.sort()

        if len(other) == 0:
            return self

        if other[0] != self.min:
            other = [self.min - 0.51] + other
        if other[-1] != self.max:
            other = other + [self.max + 0.51]

        other = [other[0]] + [other[i] for i in range(1, len(other))
                              if abs(other[i] - other[i-1]) > 0.51]

        return LogBachedDomain(
            * [NaturalDomain(other[i-1] + 0.51, other[i] - 0.51) for i in range(1, len(other))]
        )

    def __lt__(self, other):
        if is_iterable(other):
            other = min(other)

        if self.min > other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are graters that {other}")

        self.max = min(other - 0.51, self.max)
        return self

    def __gt__(self, other):
        if is_iterable(other):
            other = max(other)

        if self.max < other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are less that {other}")

        self.min = max(other + 0.51, self.min)
        return self

    def __rlt__(self, other):
        return self.__gt__(other)

    def __rgt__(self, other):
        return self.__lt__(other)

    def __ge__(self, other):
        if is_iterable(other):
            other = max(other)

        if self.max < other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are less that {other}")

        self.min = max(other, self.min)
        return self

    def __le__(self, other):
        if is_iterable(other):
            other = min(other)

        if self.min > other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are graters that {other}")

        self.max = min(other, self.max)
        return self

    def __or__(self, __o):
        return BachedDomain(self, __o)

    def _get_range(self, factor):
        if factor < 0:
            return (factor, -factor)

        return (-factor, factor)

    def __mod_eq__(self, factor, value):
        sing = factor/abs(factor)
        if is_iterable(value):

            _range = self._get_range(factor)
            value = [i for i in value if _range[0] < i and i < _range[1]]

            return nsp.New[nsp.BachedDomain](
                * [nsp.New[nsp.LinealTransformed](
                    original_domain=self,
                    transformer=lambda x: x/abs(factor),
                    inverse=lambda x: x * factor,
                    independent_value=ind * sing
                ) for ind in value]
            )

        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x/abs(factor),
            inverse=lambda x: x * factor,
            independent_value=value
        )

    def __mod_neq__(self, factor, value):
        if not is_iterable(value):
            value = [value]

        value = [abs(v) for v in value]

        return nsp.New[nsp.BachedDomain](
            * [nsp.New[nsp.LinealTransformed](
                original_domain=self,
                transformer=lambda x: x/abs(factor),
                inverse=lambda x: x * factor,
                independent_value=ind
            ) for ind in range(*self._get_range(factor))
                if not abs(ind) in value
            ]
        )

    def __mod_lt__(self, factor, value):
        if is_iterable(value):
            value = min([abs(v) for v in value])

        return nsp.New[nsp.BachedDomain](
            * [nsp.New[nsp.LinealTransformed](
                original_domain=self,
                transformer=lambda x: x/factor,
                inverse=lambda x: x * factor,
                independent_value=ind
            ) for ind in range(*self._get_range(factor)) if abs(ind) < value]
        )

    def __mod_gt__(self, factor, value):
        if is_iterable(value):
            value = max([abs(v) for v in value])

        return nsp.New[nsp.BachedDomain](
            * [nsp.New[nsp.LinealTransformed](
                original_domain=self,
                transformer=lambda x: x/factor,
                inverse=lambda x: x * factor,
                independent_value=ind
            ) for ind in range(*self._get_range(factor)) if abs(ind) > value]
        )

    def __mod_ge__(self, factor, value):
        if is_iterable(value):
            value = max([abs(v) for v in value])

        return nsp.New[nsp.BachedDomain](
            * [nsp.New[nsp.LinealTransformed](
                original_domain=self,
                transformer=lambda x: x/factor,
                inverse=lambda x: x * factor,
                independent_value=ind
            ) for ind in range(*self._get_range(factor)) if abs(ind) >= value]
        )

    def __mod_le__(self, factor, value):
        if is_iterable(value):
            value = min([abs(v) for v in value])

        return nsp.New[nsp.BachedDomain](
            * [nsp.New[nsp.LinealTransformed](
                original_domain=self,
                transformer=lambda x: x/factor,
                inverse=lambda x: x * factor,
                independent_value=ind
            ) for ind in range(*self._get_range(factor)) if abs(ind) <= value]
        )



from typing import Type, TypeVar, Generic

T = TypeVar('T')


# class Instanter(Generic[T]):
#     def __init__(self, cls) -> None:
#         super().__init__()
#         self.cls = cls

#     def __call__(self, *args, **kwg) -> T:
#         return self.cls(*args, **kwg)


class NewClass:
    def __getitem__(self, _type: Type[T]) -> Type[T]:
        return _type._class[_type]


New = NewClass()


class TypeClass:
    def __getitem__(self, _type: Type[T]) -> Type[T]:
        return _type._class[_type]


Type = TypeClass()


def default_init(cls):
    def f(self, *args, **kwg):
        super(cls,  self).__init__(*args, **kwg)
    return f


class InjectorMetaclass(type):
    _class = {}

    def __new__(cls, *args, **kwg):
        def __init__(self, *args, **kwg):
            super().__init__(*args, **kwg)
            try:
                _ = type(self)._class[type(self)]
            except KeyError:
                type(self)._class[type(self)] = self

        cls.__mro__[0].__init__ = __init__
        for i in range(1, len(cls.__mro__) - 2):
            cls.__mro__[i].__init__ = default_init(cls.__mro__[i + 1])
        return super().__new__(cls, *args, **kwg)



from .visitor_natural_ast import NaturalAstVisitor
from .visitor_domain_modifier import DomainModifierVisitor
from .visitor_sample_check import ValidateSampler
from .visitor_index_ast_modifier import IndexAstModifierVisitor
from .visitor_eval_ast_checked import EvalAstChecked
from .visitor_member_ast_modifier import MemberAstModifierVisitor
from .visitor_type_domain_modifier import TypeDomainModifierVisitor



from .__base__ import UniversalVariableBinaryOperation, check_natural_value
from search_space.utils.ast_tools import decorated_all_methods


@decorated_all_methods(check_natural_value)
class LogicOpsNode(UniversalVariableBinaryOperation):

    def __and__(self, other):
        return AndOp(self, other)

    def __rand__(self, other):
        return AndOp(other, self)

    def __or__(self, other):
        return OrOp(self, other)

    def __ror__(self, other):
        return OrOp(other, self)

    # def __xor__(self, other):
    #     return XorOp(self, other)

    # def __rxor__(self, other):
    #     return XorOp(other, self)


class AndOp(LogicOpsNode):

    def _suggestion(self, op):
        return f'try to change the constraint like {type(self.target).__name__} & ( {type(self.other).__name__} {op} ...'


class OrOp(LogicOpsNode):

    def _suggestion(self, op):
        return f'try to change the constraint like {type(self.target).__name__} | ( {type(self.other).__name__} {op} ...'



from search_space.utils.namespace_injection import InjectorMetaclass, New, Type


class ExprNode(InjectorMetaclass):
    def __init__(self, target, other):
        pass


class NaturalValueNode(ExprNode):
    def __init__(self, value):
        pass



from copy import copy
from search_space.errors import ArgumentFunctionError, NotEvaluateError
from search_space.spaces.asts.naturals_values.__base__ import NaturalValuesNode
from search_space.utils import visitor
from search_space.spaces.asts import constraints
from . import VisitorLayer
from .visitor_natural_ast import NaturalAstVisitor


class TypeDomainModifierVisitor(VisitorLayer):
    def __init__(self) -> None:
        super().__init__()
        self.natural_visitor = NaturalAstVisitor()

    @property
    def do_transform_to_check_sample(self):
        return False

    def domain_optimization(self, node, domain):
        self.domain, self._context = domain, None
        return self.visit(node)

    def transform_to_modifier(self, node, domain, context):
        self.domain, self._context = domain, context
        return self.visit(node)

    def __op_apply__(self, a, b, op):
        if visitor.WithoutDispatcher in [type(a), type(b)]:
            return visitor.without_dispatcher

        if a is None:
            if b is None or isinstance(b, type):
                self.domain = op(self.domain, b)
            return None

        if b is None:
            if isinstance(b, type):
                self.domain = op(a, self.domain)
            return None

        return op(b, a)

    @visitor.on("node")
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node: constraints.AstRoot):

        for n in node.asts:
            self.visit(n)

        return node, self.domain

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AndOp)
    def visit(self, node):
        a = self.visit(node.target)

        if type(a) == bool and not a:
            return

        b = self.visit(node.other)

        return self.__op_apply__(a, b, lambda a, b: a & b)

    @visitor.when(constraints.OrOp)
    def visit(self, node):
        current_domain = copy(self.domain)
        a = self.visit(node.target)

        if type(a) == bool and a:
            return

        current_domain, self.domain = self.domain, current_domain
        b = self.visit(node.other)

        if type(a) == bool and b is None:
            return

        return self.__op_apply__(a, b, lambda a, b: a | b)

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.LessOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x < y)

    @visitor.when(constraints.LessOrEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x <= y)

    @visitor.when(constraints.GreatOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x > y)

    @visitor.when(constraints.GreatOrEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x >= y)

    @visitor.when(constraints.NotEqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x != y)

    @visitor.when(constraints.EqualOp)
    def visit(self, node):
        target = self.visit(node.target)
        limit = self.visit(node.other)

        return self.__op_apply__(target, limit, lambda x, y: x == y)

    @visitor.when(constraints.SelfNode)
    def visit(self, node):
        pass

    @visitor.when(constraints.NaturalValue)
    def visit(self, node):
        if isinstance(node.target, NaturalValuesNode):
            return self.natural_visitor.get_value(
                node.target, context=self._context
            )

        return node.target

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg)

        return node.func(*new_args, **new_kw)

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.FunctionNode):
        raise ArgumentFunctionError()

    @visitor.when(constraints.NotEvaluate)
    def visit(self, node):
        raise NotEvaluateError()



# -*- coding: utf8 -*-

__author__ = 'Suilan Estévez Velarde'


INFINITY_MAX_VALUE = 100000000000000000000
INFINITY_MIN_VALUE = - 100000000000000000000


class Infinity:
    """Un valor mayor que todos los valores posibles
    de int() para comparaciones en el minimax
    """

    def __init__(self, positive=True):
        self.positive = positive

    def __le__(self, other):
        return not self.positive

    def __ge__(self, other):
        return self.positive

    def __lt__(self, other):
        return not self.positive

    def __gt__(self, other):
        return self.positive

    def __eq__(self, other):
        return False

    def __neg__(self):
        return Infinity(not self.positive)

    def __nonzero__(self):
        return True

    def __float__(self):
        if self.positive:
            return float(INFINITY_MAX_VALUE)

        return float(INFINITY_MIN_VALUE)

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return -self

    def __mul__(self, other):
        if other == 0:
            return 0

        if other < 0:
            return -self

        return self

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if other == 0:
            raise "Can't divide for zero"

        if other < 0:
            return -self

        return self

    def __rtruediv__(self, other):
        return 0

    def __str__(self) -> str:
        if self.positive:
            return 'oo'

        return '-oo'

    def __repr__(self) -> str:
        return self.__str__()

    def __abs__(self):
        if self.positive:
            return float(INFINITY_MAX_VALUE)

        return float(INFINITY_MIN_VALUE)


# Instancia global de infinito
oo = Infinity()



from search_space.errors import UnSupportOpError
from search_space.utils.ast_tools import check_params_type
from search_space.utils.itertools import is_iterable
from . import __namespace__ as nsp
from typing import Iterable


class UniversalVariableNode:
    precedence = 1

    @property
    def is_self(self):
        return False

    def _suggestion(self, op):
        return ''

    #################################################################
    #                                                               #
    #                     Class Operations                          #
    #                                                               #
    #################################################################

    def __getattr__(self, name):
        raise UnSupportOpError(self, name, 'getattr',
                               self._suggestion('getattr'))

    #################################################################
    #                                                               #
    #                     List Operations                           #
    #                                                               #
    #################################################################

    def __getitem__(self, index):
        raise UnSupportOpError(self, index, 'getitem',
                               self._suggestion('getitem'))

    #################################################################
    #                                                               #
    #                     Unary Operations                          #
    #                                                               #
    #################################################################

    def __neg__(self):
        raise UnSupportOpError(self, None, '-',
                               self._suggestion('-'))

    #################################################################
    #                                                               #
    #              Binary Arithmetic Operations                     #
    #                                                               #
    #################################################################

    def __add__(self, other):
        raise UnSupportOpError(self, other, '+',
                               self._suggestion('+'))

    def __radd__(self, other):
        raise UnSupportOpError(self, other, '+',
                               self._suggestion('+'))

    def __mul__(self, other):
        raise UnSupportOpError(self, other, '*',
                               self._suggestion('*'))

    def __rmul__(self, other):
        raise UnSupportOpError(self, other, '*',
                               self._suggestion('*'))

    def __sub__(self, other):
        raise UnSupportOpError(self, other, '-',
                               self._suggestion('_'))

    def __rsub__(self, other):
        raise UnSupportOpError(self, other, '-',
                               self._suggestion('-'))

    def __div__(self, other):
        raise UnSupportOpError(self, other, '/',
                               self._suggestion('/'))

    def __rdiv__(self, other):
        raise UnSupportOpError(self, other, '/',
                               self._suggestion('/'))

    def __mod__(self, other):
        raise UnSupportOpError(self, other, '%',
                               self._suggestion('%'))

    def __divmod__(self, other):
        raise UnSupportOpError(self, other, '%',
                               self._suggestion('%'))

    #################################################################
    #                                                               #
    #                 Binary Logical Operations                     #
    #                                                               #
    #################################################################

    def __and__(self, other):
        raise UnSupportOpError(self, other, '&',
                               self._suggestion('&'))

    def __rand__(self, other):
        raise UnSupportOpError(self, other, '&',
                               self._suggestion('&'))

    def __or__(self, other):
        raise UnSupportOpError(self, other, '|',
                               self._suggestion('|'))

    def __ror__(self, other):
        raise UnSupportOpError(self, other, '|',
                               self._suggestion('|'))

    # def __xor__(self, other):
    #     return XorOp(self, other)

    # def __rxor__(self, other):
    #     return XorOp(other, self)

    #################################################################
    #                                                               #
    #                 Binary Compare Operations                     #
    #                                                               #
    #################################################################

    def __eq__(self, other):
        raise UnSupportOpError(self, other, '==',
                               self._suggestion('=='))

    def __req__(self, other):
        raise UnSupportOpError(self, other, '==',
                               self._suggestion('=='))

    def __ne__(self, other):
        raise UnSupportOpError(self, other, '!=',
                               self._suggestion('!='))

    def __rne__(self, other):
        raise UnSupportOpError(self, other, '!=',
                               self._suggestion('!='))

    def __lt__(self, other):
        raise UnSupportOpError(self, other, '<',
                               self._suggestion('<'))

    def __rlt__(self, other):
        raise UnSupportOpError(self, other, '<',
                               self._suggestion('<'))

    def __gt__(self, other):
        raise UnSupportOpError(self, other, '>',
                               self._suggestion('>'))

    def __rgt__(self, other):
        raise UnSupportOpError(self, other, '>',
                               self._suggestion('>'))

    def __ge__(self, other):
        raise UnSupportOpError(self, other, '>=',
                               self._suggestion('>='))

    def __rge__(self, other):
        raise UnSupportOpError(self, other, '>=',
                               self._suggestion('>='))

    def __le__(self, other):
        raise UnSupportOpError(self, other, '<=',
                               self._suggestion('<='))

    def __rle__(self, other):
        raise UnSupportOpError(self, other, '<=',
                               self._suggestion('<='))


class UniversalVariableBinaryOperation(UniversalVariableNode):
    def __init__(self, target, other) -> None:
        super().__init__()
        self.target = target
        self.other = other

    def inverted_op(self, a, b):
        return type.__call__(self.__class__, a, b)

    def reverse(self):
        self.other, self.target = self.target, self.other
        return self

    def op(self):
        pass


class AstRoot(UniversalVariableNode):
    def __init__(self, asts=None) -> None:
        super().__init__()
        self.asts = [] if asts is None else asts

    def add_constraint(self, ast):
        if is_iterable(ast):
            self.asts += ast
        else:
            self.asts.append(ast)

    def __add__(self, other):
        if other is None:
            return self

        n = AstRoot([])

        if is_iterable(other):
            n.asts = self.asts + other
        else:
            n.asts = self.asts + other.asts

        return n


check_natural_value = check_params_type(
    lambda: UniversalVariableNode,
    lambda x: nsp.New[nsp.NaturalValueNode](x)
)



from . import __namespace__ as nsp


class Domain(nsp.DomainProtocol):
    def is_invalid(self):
        return False

    def __rlt__(self, other):
        return self.__gt__(other)

    def __rgt__(self, other):
        return self.__lt__(other)

    def __rle__(self, other):
        return self.__ge__(other)

    def __rge__(self, other):
        return self.__le__(other)

    def __or__(self, __o):
        return nsp.New[nsp.BachedDomain](self, __o)


class NumeralDomain(Domain, nsp.NumeralDomainProtocol):
    def is_invalid(self):
        return self.min > self.max

    def extend(self, top):
        if self.max < top:
            self.max = top

    def __add__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x - factor,
            inverse=lambda x: x,
            independent_value=0
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x + factor,
            inverse=lambda x: x,
            independent_value=0
        )

    def __rsub__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x,
            inverse=lambda x: factor - x,
            independent_value=0
        )

    def __mul__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x / factor,
            inverse=lambda x: x,
            independent_value=0
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x * factor,
            inverse=lambda x: x,
            independent_value=0
        )

    def __rtruediv__(self, factor):
        return nsp.New[nsp.LinealTransformed](
            original_domain=self,
            transformer=lambda x: x * factor,
            inverse=lambda x: x / factor,
            independent_value=0
        )

    def __contains__(self, value):
        return self.min <= value and value <= self.max



from .__base__ import UniversalVariableNode, AstRoot, UniversalVariableBinaryOperation
from .logic_ops import AndOp, OrOp
from .cmp_ops import EqualOp, NotEqualOp, LessOp, LessOrEqualOp, GreatOp, GreatOrEqualOp
from .expr_ops import AddOp, SubOp, MultiOp, DivOp
from .segmentation_branch import SegmentationAddOp, SegmentationSubOp, SegmentationMultiOp, SegmentationDivOp
from .segmentation_branch import SegmentationEqualOp, SegmentationNotEqualOp, SegmentationLessOp, SegmentationLessOrEqualOp, SegmentationGreatOp, SegmentationGreatOrEqualOp
from .segmentation_branch import SegmentationModOp, SegmentationExprNode
from .atoms import GetAttr, GetItem, NaturalValue, NotEvaluate, SelfNode, FunctionNode, AdvancedFunctionNode

"""
         Constraint := Cmp | Constraint
                    := Cmp & Constraint
                    := Cmp
                    := SegmentationExpr
 
                Cmp := Expr == Expr
                Cmp := Expr != Expr
                Cmp := Expr >= Expr
                Cmp := Expr > Expr
                Cmp := Expr <= Expr
                Cmp := Expr < Expr
                Cmp := Expr

   SegmentationExpr := SegmentationAtom == Expr
   SegmentationExpr := SegmentationAtom != Expr
   SegmentationExpr := SegmentationAtom <= Expr
   SegmentationExpr := SegmentationAtom < Expr
   SegmentationExpr := SegmentationAtom > Expr
   SegmentationExpr := SegmentationAtom >= Expr

   SegmentationAtom := SegmentationAtom + Expr  
   SegmentationAtom := SegmentationAtom - Expr  
   SegmentationAtom := SegmentationAtom * Expr  
   SegmentationAtom := SegmentationAtom / Expr  
   SegmentationAtom := SegmentationAtom % Expr  
   SegmentationAtom := Expr % Expr 


               Expr := Atom + Expr
               Expr := Atom - Expr
               Expr := Atom * Expr
               Expr := Atom / Expr
               Expr := Atom
     
               Atom := Atom[index]
               Atom := Atom[member]
               Atom := Self
               Atom := (Constraint)
               Atom := NaturalValue

       NaturalValue := FunctionCall
       NaturalValue := NaturalValuesAst
       NaturalValue := int, str, float, list .....

"""



from copy import copy
from typing import List
from search_space.context_manager.runtime_manager import SearchSpaceConfig
from search_space.errors import InvalidSpaceDefinition
from search_space.spaces.domains.__base__ import Domain
from search_space.sampler import Sampler
from search_space.sampler.distribution_names import UNIFORM
from . import __namespace__ as nsp
from search_space.spaces import domains


class BachedDomain(Domain, metaclass=nsp.BachedDomain):
    def __init__(self, *domains: List[Domain]) -> None:
        self.domains: List[Domain] = list(domains)
        self.sampler_selector: Sampler = None

    def extend(self, top):
        self.domains[-1].extend(top)

    def is_invalid(self):
        index = 0
        while index < len(self.domains):

            if self.domains[index].is_invalid():
                self.domains.pop(index)
            else:
                index += 1

        return len(self.domains) == 0

    @property
    def limits(self):
        self.is_invalid()
        return tuple([d.limits for d in self.domains])

    def get_sample(self, sampler: Sampler):
        if self.sampler_selector is None:
            self.sampler_selector = SearchSpaceConfig(
            ).sampler_manager.create_sampler(UNIFORM, self)
        bache = self.sampler_selector.choice(self.domains)
        return bache.get_sample(sampler)

    def __copy__(self):
        return BachedDomain(*[copy(d) for d in self.domains])

    def _propagation(self, f):
        new_space = []
        for domain in self.domains:
            try:
                domain = f(domain)

            except InvalidSpaceDefinition:
                continue

            if isinstance(domain, BachedDomain):
                new_space += [d for d in domain.domains if not d.is_invalid()]
            else:
                new_space.append(domain)

        self.domains = new_space

    def _find_bache(self, n):
        r, l = 0, len(self.domains)
        m = l + r >> 1

        while not n in self.domains[m] and r < l:
            _min, _max = self.domains[m].limits

            if n > _max:
                r = m + 1
            elif n < _min:
                l = m - 1

            m = l + r >> 1

        return m

    def __eq__(self, other):
        self._propagation(lambda x: x == other)
        return self

    def __ne__(self, other):
        # index = self._find_bache(other)
        # domain = self.domains[index] != other

        # if isinstance(domain, BachedDomain):
        #     self.domains = (
        #         self.domains[:index] +
        #         [d for d in domain.domains if not d.is_invalid()] +
        #         self.domains[index + 1:]
        #     )
        self._propagation(lambda x: x != other)

        return self

    def __lt__(self, other):
        self._propagation(lambda x: x < other)
        return self

    def __gt__(self, other):
        self._propagation(lambda x: x > other)
        return self

    def __ge__(self, other):
        self._propagation(lambda x: x >= other)
        return self

    def __le__(self, other):
        self._propagation(lambda x: x <= other)
        return self

    def __or__(self, __o):
        return self.domains.append(__o)

    def __hash__(self) -> int:
        return id(self)


class LogBachedDomain(BachedDomain, metaclass=nsp.LogBachedDomain):
    def __init__(self, *domains: List[Domain]) -> None:
        super().__init__(*domains)

    def __copy__(self):
        return LogBachedDomain(*[copy(d) for d in self.domains])

    def _find_bache(self, n):
        r, l = 0, len(self.domains)
        m = l + r >> 1

        while r < l and not n in self.domains[m]:
            _min, _max = self.domains[m].limits

            if n > _max:
                r = m + 1
            elif n < _min:
                l = m - 1

            m = l + r >> 1

        return m if m < len(self.domains) else len(self.domains) - 1

    def __eq__(self, other):
        index = self._find_bache(other)
        return self.domains[index] == other

    def __ne__(self, other):
        index = self._find_bache(other)
        domain = self.domains[index] != other

        if isinstance(domain, BachedDomain):
            self.domains = (
                self.domains[:index] +
                [d for d in domain.domains if not d.is_invalid()] +
                self.domains[index + 1:]
            )

        return self

    def __lt__(self, other):
        index = self._find_bache(other)
        self.domains = self.domains[:index] + [self.domains[index] < other]

        return self

    def __gt__(self, other):
        index = self._find_bache(other)
        self.domains = [self.domains[index] > other] + self.domains[index + 1:]
        return self

    def __ge__(self, other):
        index = self._find_bache(other)
        self.domains = [self.domains[index] >=
                        other] + self.domains[index + 1:]
        return self

    def __le__(self, other):
        index = self._find_bache(other)
        self.domains = self.domains[:index] + [self.domains[index] <= other]

        return self

    def __or__(self, __o):
        bd = BachedDomain(*self.domains) | __o
        bd.sample_selector = self.sampler_selector
        return bd

    def __hash__(self) -> int:
        return id(self)



from search_space.errors import InvalidSpaceConstraint, InvalidSpaceDefinition
from search_space.spaces.domains.__base__ import Domain
from .bached_domain import BachedDomain
from search_space.sampler import Sampler
from typing import Iterable
from search_space.utils.itertools import is_iterable


class CategoricalDomain(Domain):
    def __init__(self, _list) -> None:
        self.list = _list
        self.maps = []

    def is_invalid(self):
        return len(self.list) == 0

    def __copy__(self):
        return CategoricalDomain(self.list)

    def get_sample(self, sampler: Sampler):
        return sampler.choice(self.list)

    @property
    def limits(self):
        return self.list

    def filter(self, predicate):
        result = []
        for i, item in enumerate(self.list):
            for _map in self.maps:
                item = _map(item)

            if predicate(item):
                result.append(self.list[i])

        self.list = result

    #################################################################
    #                                                               #
    #                     List Operations                           #
    #                                                               #
    #################################################################

    def __getitem__(self, index):
        raise InvalidSpaceConstraint("Numeral value isn't indexed")

    #################################################################
    #                                                               #
    #                 Binary Compare Operations                     #
    #                                                               #
    #################################################################

    def __eq__(self, other):
        if is_iterable(other) and not isinstance(other, str):
            self.filter(lambda x: x in other)
            return self

        self.filter(lambda x: x == other)
        return self

    def __ne__(self, other):
        if is_iterable(other) and not isinstance(other, str):
            self.filter(lambda x: not x in other)
            return self

        self.filter(lambda x: x != other)
        return self

    def __lt__(self, other):
        if is_iterable(other):
            other = min(other)

        self.filter(lambda x: x < other)
        return self

    def __gt__(self, other):
        if is_iterable(other) and not isinstance(other, str):
            other = max(other)
        self.filter(lambda x: x > other)
        return self

    def __ge__(self, other):
        if is_iterable(other) and not isinstance(other, str):
            other = max(other)

        self.filter(lambda x: x >= other)
        return self

    def __le__(self, other):
        if is_iterable(other):
            other = min(other)

        self.filter(lambda x: x <= other)
        return self

    def __or__(self, __o):
        return BachedDomain(self, __o)



from ast import arg


def decorated_all_methods(decorator):
    def f(cls):
        for attr in cls.__dict__:  # there's propably a better way to do this
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return f


def check_ast_precedence(fun):
    def f(*args, **kws):
        if len(args) == 2:

            _self = args[0]
            other = args[1]

            try:
                p = other.precedence
            except AttributeError:
                p = -1

            if p > _self.precedence:

                f = getattr(other, fun.__name__)
                return f(_self).reverse()
        return fun(*args, **kws)
    f.__name__ = fun.__name__
    return f


def check_params_type(typ_fun, class_fun):
    def ff(func):

        def f(*args, **kws):
            new_args = []
            for arg in args:
                if isinstance(arg, typ_fun()):
                    new_args.append(arg)
                else:
                    new_args.append(class_fun(arg))
            return func(*new_args, **kws)
        f.__name__ = func.__name__
        return f
    return ff


def index_list(shape):
    dp = [[i] for i in range(shape[0])]

    for dim in range(1, len(shape)):
        dp = [current_index + [i] for current_index in dp for i in range(dim)]

    return dp



from search_space.errors import UnSupportOpError
from .logic_ops import LogicOpsNode
from .__base__ import check_natural_value
from search_space.utils.ast_tools import decorated_all_methods
from . import __namespace__ as nsp


def expr_node_check(fn):
    def f(*args):
        if not isinstance(args[1], nsp.Type[nsp.ExprNode]):
            raise UnSupportOpError(args[0], args[1], 'cmp')

        return fn(*args)
    f.__name__ = fn.__name__
    return f


@decorated_all_methods(check_natural_value)
@decorated_all_methods(expr_node_check)
class CmpOpsNode(LogicOpsNode):

    def __eq__(self, other):
        return EqualOp(self, other)

    def __req__(self, other):
        return EqualOp(other, self)

    def __ne__(self, other):
        return NotEqualOp(self, other)

    def __rne__(self, other):
        return NotEqualOp(other, self)

    def __lt__(self, other):
        return LessOp(self, other)

    def __rlt__(self, other):
        return LessOp(other, self)

    def __gt__(self, other):
        return GreatOp(self, other)

    def __rgt__(self, other):
        return GreatOp(other, self)

    def __ge__(self, other):
        return GreatOrEqualOp(self, other)

    def __rge__(self, other):
        return GreatOrEqualOp(other, self)

    def __le__(self, other):
        return LessOrEqualOp(self, other)

    def __rle__(self, other):
        return LessOrEqualOp(other, self)

    @staticmethod
    def setter_expr_class(cls):
        CmpOpsNode.ExprNode = cls
        return cls


class EqualOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} == ( {type(self.other).__name__} {op} ..."""

    def op(self):
        return lambda x, y: x == y


class NotEqualOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} != ( {type(self.other).__name__} {op} ..."""

    def op(self):
        return lambda x, y: x != y


class GreatOrEqualOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} >= ( {type(self.other).__name__} {op} ..."""

    def inverted_op(self, a, b):
        return LessOrEqualOp(a, b)

    def op(self):
        return lambda x, y: x >= y


class GreatOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} > ( {type(self.other).__name__} {op} ..."""

    def inverted_op(self, a, b):
        return LessOp(a, b)

    def op(self):
        return lambda x, y: x > y


class LessOrEqualOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} <= ( {type(self.other).__name__} {op} ..."""

    def inverted_op(self, a, b):
        return GreatOrEqualOp(a, b)

    def op(self):
        return lambda x, y: x <= y


class LessOp(LogicOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
            {type(self.target).__name__} > ( {type(self.other).__name__} {op} ..."""

    def inverted_op(self, a, b):
        return GreatOp(a, b)

    def op(self):
        return lambda x, y: x < y



from typing import Protocol, Union
from search_space.utils.singleton import Singleton
from search_space.utils import infinity
from search_space.sampler import SamplerFactory


class SearchSpacePrinter:

    def __init__(self) -> None:
        self.tabs = 0

    def init_search(self, id_space, name_space):
        pass

    def sample_value(self, value, caching_value=False):
        pass

    def domain_init(self, domain):
        pass

    def ast_transformation(self, domain, ast, visitor_name=''):
        pass

    def sample_error(self, sample, error, sample_num, domain):
        pass

    def context_name(self, context):
        pass


class SearchSpaceConfig(metaclass=Singleton):
    def __init__(
        self,
        verbose: bool = False,
        replay_nums: Union[int, None] = None,
        ast_optimizations:  bool = True,
        printer: SearchSpacePrinter = None,
        minimal_numeral_limit=None,
        maximal_numeral_limit=None,
        sampler_manager: SamplerFactory = SamplerFactory
    ) -> None:
        self.verbose = verbose
        self.replay_nums = replay_nums
        self.ast_optimizations = ast_optimizations
        self.printer = printer
        self.attempts = []
        self.sampler_manager = sampler_manager()

        if minimal_numeral_limit:
            infinity.INFINITY_MIN_VALUE = minimal_numeral_limit

        if maximal_numeral_limit:
            infinity.INFINITY_MAX_VALUE = maximal_numeral_limit

    @property
    def printer_class(self) -> SearchSpacePrinter:
        return SearchSpacePrinter() if not self.verbose else self.printer



class SamplerInfo:
    def __init__(self, scope, domain, sampler_value, real_value) -> None:
        self.scope = scope
        self.domain = domain
        self.sampler_value = sampler_value
        self.real_value = real_value

    def __str__(self) -> str:
        return f'{self.scope.scope}[{self.domain}] ---> {self.sampler_value} ---> {self.real_value}'


class ConstraintInfo:
    def __init__(self, scope, constraint_name, initial_domain, result) -> None:
        self.scope = scope
        self.initial_domain = initial_domain
        self.result = result
        self.constraint_name = constraint_name

    def __str__(self) -> str:
        return f'{self.scope}[{self.constraint_name}]: {self.initial_domain} ---> {self.result}'


class InitSamplerInfo:
    def __init__(self, scope, domain) -> None:
        self.scope = scope
        self.domain = domain

    def __str__(self) -> str:
        return f'Sampler in {self.scope.scope}[{self.domain}]:'


class SamplerContext:
    def __init__(self, name='', father=None) -> None:
        self.name = name
        self.sampler_logs = []
        self.context = {}
        self.result = None
        self.father: SamplerContext = father
        self.sampling_status = {}

    #################################################################
    #                                                               #
    #                     Sampler Registry                          #
    #                                                               #
    #################################################################

    def create_child(self, name=''):
        return SamplerContext(name=name, father=self)

    def registry_sampler(self, search_space, value):
        self.context[hash(search_space)] = value

    def registry_init_sampler_process(self, search_space):
        self.sampling_status[hash(search_space)] = True

    def check_sampling_status(self, search_space):
        try:
            return self.sampling_status[hash(search_space)]
        except KeyError:
            if not self.father is None:
                return self.father.check_sampling_status(search_space)
            return None

    def get_sampler_value(self, search_space):
        try:
            return self.context[hash(search_space)]
        except KeyError:
            if not self.father is None:
                return self.father.get_sampler_value(search_space)
            return None

    #################################################################
    #                                                               #
    #                     Sampler Logs                              #
    #                                                               #
    #################################################################

    def push_log(self, log):
        self.sampler_logs.append(log)

    def print_history(self):
        result, index = [], 0
        while index < len(self.sampler_logs):
            r, index = self.__print_history(index=index)
            result += r

        print('\n'.join(result))

    def __print_history(self, index=0, tabs=''):
        result = [tabs + str(self.sampler_logs[index])]
        index += 1
        while index < len(self.sampler_logs):
            log = self.sampler_logs[index]
            if isinstance(log, InitSamplerInfo):
                r, index = self.__print_history(index, tabs + '\t')
                result += r
            else:
                result.append(tabs + '\t' + str(log))
                if isinstance(log, SamplerInfo):
                    return result, index + 1
                index += 1
        return result, index



from search_space.utils import visitor
from search_space.spaces.asts import constraints
from search_space.errors import InvalidSampler, NotEvaluateError
from search_space.utils.ast_tools import index_list
from . import VisitorLayer
from search_space.utils.singleton import Singleton
from .visitor_natural_ast import NaturalAstVisitor, NaturalValuesNode
from typing import Iterable
from search_space.utils.itertools import is_iterable


class ValidateSampler(VisitorLayer, metaclass=Singleton):

    def __init__(self) -> None:
        self.natural_visitor = NaturalAstVisitor()

    def check_is_not_eval_node(self, *nodes):
        for node in nodes:
            try:
                if not node.can_evaluate:
                    return node
            except AttributeError:
                pass
        return None

    @property
    def do_transform_to_modifier(self):
        return False

    def transform_to_check_sample(self, node, sample, context):
        self.sample, self.context = sample, context
        return self.visit(node)

    def check_sample_by_index(self, node, sample, context, index):
        self.current_index = index
        return self.transform_to_check_sample(node, sample, context)

    @visitor.on("node")
    def visit(self, node, current_index=[], is_main_node=False):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node, current_index=[], is_main_node=False):
        for n in node.asts:
            try:
                _ = self.visit(n, current_index, is_main_node=True)
            except NotEvaluateError:
                pass

    #################################################################
    #                                                               #
    #                 Arithmetic Visit                              #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AddOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a + b

    @visitor.when(constraints.SubOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a - b

    @visitor.when(constraints.MultiOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a * b

    @visitor.when(constraints.DivOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        return a / b

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(constraints.AndOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = self.visit(node.target, current_index)

        if type(a) == bool and not a:
            return a

        return self.visit(node.other, current_index)

    @visitor.when(constraints.OrOp)
    def visit(self, node, current_index=[], is_main_node=False):
        a = False
        try:
            a = self.visit(node.target, current_index)
        except InvalidSampler:
            pass

        if a:
            return a

        return self.visit(node.other, current_index)

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.NotEqualOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b) and not is_iterable(a):
            if not a in b:
                return self

        elif is_iterable(a) and not is_iterable(b):
            if not b in a:
                return self

        elif a != b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} != {b} ]")

    @visitor.when(constraints.EqualOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b) and not is_iterable(a):
            if a in b:
                return self

        elif is_iterable(a) and not is_iterable(b):
            if b in a:
                return self

        elif a == b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} == {b} ]")

    @visitor.when(constraints.LessOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = min(b)

        if is_iterable(a):
            a = max(a)

        if a < b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} < {b} ]")

    @visitor.when(constraints.LessOrEqualOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = min(b)

        if is_iterable(a):
            a = max(a)

        if a <= b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} < {b} ]")

    @visitor.when(constraints.GreatOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = max(b)

        if is_iterable(a):
            a = min(a)

        if a > b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} > {b} ]")

    @visitor.when(constraints.GreatOrEqualOp)
    def visit(self, node, current_index=[], is_main_node=False):

        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if is_iterable(b):
            b = max(b)

        if is_iterable(a):
            a = min(a)

        if a >= b:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {a} >= {b} ]")

    #################################################################
    #                                                               #
    #                 Segmentation Visit                            #
    #                                                               #
    #################################################################

    @visitor.when(constraints.SegmentationEqualOp)
    def visit(self, node: constraints.SegmentationEqualOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        mod = abs(target % factor)
        if is_iterable(value) and mod in value:
            return self

        if mod == abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} == {value} ]")

    @visitor.when(constraints.SegmentationNotEqualOp)
    def visit(self, node: constraints.SegmentationNotEqualOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        mod = target % factor
        if (
            is_iterable(value)
            and not mod in value
            and not -mod in value
        ):
            return self

        if mod != value and -mod != value:
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} != {value} ]")

    @visitor.when(constraints.SegmentationLessOp)
    def visit(self, node: constraints.SegmentationLessOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = min(value)

        if abs(target % factor) < abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} < {value} ]")

    @visitor.when(constraints.SegmentationLessOrEqualOp)
    def visit(self, node: constraints.SegmentationLessOrEqualOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = min(value)

        if abs(target % factor) <= abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} <= {value} ]")

    @visitor.when(constraints.SegmentationGreatOp)
    def visit(self, node: constraints.SegmentationGreatOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = max(value)

        if abs(target % factor) > abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} > {value} ]")

    @visitor.when(constraints.SegmentationGreatOrEqualOp)
    def visit(self, node: constraints.SegmentationGreatOrEqualOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)
        value = self.visit(node.value, current_index)

        if is_iterable(value):
            value = max(value)

        if abs(target % factor) >= abs(value):
            return self

        raise InvalidSampler(
            f"inconsistent sampler => [ {target} % {factor} >= {value} ]")

    @visitor.when(constraints.SegmentationAddOp)
    def visit(self, node: constraints.SegmentationAddOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.value, current_index)

        if not target is None:
            return target + factor

        self.domain = self.domain.__mod_add__(factor, 0)

    @visitor.when(constraints.SegmentationSubOp)
    def visit(self, node: constraints.SegmentationSubOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)

        if not target is None:
            return target - factor

        self.domain = self.domain.__mod_sub__(factor, 0)

    @visitor.when(constraints.SegmentationMultiOp)
    def visit(self, node: constraints.SegmentationMultiOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)

        if not target is None:
            return target * factor

        self.domain = self.domain.__mod_mult__(factor, 0)

    @visitor.when(constraints.SegmentationDivOp)
    def visit(self, node: constraints.SegmentationDivOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)

        if not target is None:
            return target / factor

        self.domain = self.domain.__mod_div__(factor, 0)

    @visitor.when(constraints.SegmentationModOp)
    def visit(self, node: constraints.SegmentationModOp, current_index=[], is_main_node=False):
        target = self.visit(node.target, current_index)
        factor = self.visit(node.other, current_index)

        if not target is None:
            return target % factor

        self.domain = self.domain.__mod_eq__(factor, 0)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.GetItem)
    def visit(self, node: constraints.GetItem, current_index=[], is_main_node=False):

        b = self.visit(node.other, current_index)
        if type(b) == bool:
            b = self.current_index[-len(current_index) - 1]

        a = self.visit(node.target, current_index + [b])

        try:
            return a[b]
        except IndexError:
            raise NotEvaluateError()
        except TypeError:
            raise InvalidSampler(
                f"inconsistent sampler => sample isn't indexed")

    @visitor.when(constraints.GetAttr)
    def visit(self, node: constraints.GetAttr, current_index=[], is_main_node=False):

        b = self.visit(node.other, current_index)
        a = self.visit(node.target, current_index)

        try:
            space = getattr(a, b)
        except AttributeError:
            raise InvalidSampler(
                f"inconsistent sampler => {a} don't has {b} member")

        try:
            sampler = space.get_sample
        except AttributeError:
            raise InvalidSampler(
                f"inconsistent sampler => {a}.{b} isn't search space")

        return sampler(context=a.__instance_context__)[0]

    @ visitor.when(constraints.SelfNode)
    def visit(self, node, current_index=[], is_main_node=False):
        return self.sample

    @ visitor.when(constraints.NaturalValue)
    def visit(self, node, current_index=[], is_main_node=False):
        if isinstance(node.target, NaturalValuesNode):
            try:
                index = self.current_index
            except AttributeError:
                index = []
            return self.natural_visitor.get_value(
                node.target,
                context=self.context,
                current_index=index
            )
        return node.target

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode, current_index=[], is_main_node=False):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg, current_index))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg, current_index)

        result = node.func(*new_args, **new_kw)

        if not result and is_main_node:
            raise InvalidSampler("False Restriction")

        return result

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.FunctionNode, current_index=[], is_main_node=False):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg, current_index))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg, current_index)

        result = node.func(*new_args, **new_kw)
        if not result and is_main_node:
            raise InvalidSampler("False Restriction")
        return result



"""
A is a Python framework for describing search spaces and generating samples from them.

The tool defines a series of design patterns with which language classes can be 
decorated by expressing in detail each of the domains relative to the class parameters 
of those classes. Using the name of any previously decorated classes, the system offers 
the necessary mechanisms to generate random instances of the described space.
"""

# The [`sampler`](ref:search_space.sampler.__init__) submodule allows us to
# define the protocols for random number generation and selection of options.
from search_space import sampler

from search_space import context_manager

from search_space.dsl import Domain, RandomValue

from search_space import errors

from search_space import functions

from search_space import spaces

from search_space import utils



from copy import copy
from distutils.command.config import config
from typing import List

from search_space.context_manager.runtime_manager import SearchSpaceConfig
from search_space.sampler import Sampler
from search_space.sampler.distribution_names import UNIFORM
from search_space.context_manager import SamplerContext
from search_space.errors import ArgumentFunctionError, DetectedRuntimeDependency, InvalidSampler, InvalidSpaceConstraint, InvalidSpaceDefinition, NotEvaluateError, CircularDependencyDetected, UndefinedSampler
from .asts import constraints as ast_constraint
from .asts import naturals_values as ast_natural
from .visitors import visitors
from .visitors import VisitorLayer
from .printer_tools.default_printer_class import DefaultPrinter

import inspect


class BasicSearchSpace:

    #################################################################
    #                                                               #
    #                     Space Initialize                          #
    #                                                               #
    #################################################################

    def __init__(self, initial_domain, distribute_like=UNIFORM, sampler=None, name=None) -> None:
        # super().__init__()
        self.initial_domain = initial_domain
        self.__distribute_like__: str = distribute_like

        self.visitor_layers: List[VisitorLayer] = [
            visitors.DomainModifierVisitor()]
        self._inner_hash = None
        self._clean_asts = ast_constraint.AstRoot([])
        self.space_name = self.__class__.__name__ if name is None else name

        self._distribution: Sampler = SearchSpaceConfig().sampler_manager.create_sampler(
            self.__distribute_like__, self) if sampler is None else sampler

    def change_distribution(self, distribution):
        self.__distribute_like__ = distribution
        self._distribution = SearchSpaceConfig().sampler_manager.create_sampler(
            self.__distribute_like__, self)

    def layers_append(self, *args):
        self.visitor_layers += list(args)

    def set_hash(self, v):
        self._inner_hash = v

    def __hash__(self) -> int:
        if self._inner_hash is None:
            return id(self)

        return self._inner_hash

    def __copy__(self):
        try:
            domain = self.initial_domain.limits

        except AttributeError:
            domain = self.initial_domain,

        result = type(self)(
            *domain, distribute_like=self.__distribute_like__)
        result.initial_domain = copy(self.initial_domain)
        result.space_name = f"{result.space_name}'"
        result.visitor_layers = [item for item in self.visitor_layers]
        result._clean_asts.add_constraint(self._clean_asts.asts)
        return result

    #################################################################
    #                                                               #
    #                     Ast Optimization                          #
    #                                                               #
    #################################################################

    def __ast_optimization__(self, ast_list):
        """
        """

        ast = ast_constraint.AstRoot([])

        if type(ast_list) == type(ast):
            temp_ast = ast_list
        else:

            if callable(ast_list):
                ast_list = [ast_list]

            temp_ast = ast_constraint.AstRoot([])
            for func in ast_list:
                new_ast = self.__build_constraint__(func)
                temp_ast.add_constraint(new_ast)

        for new_ast in temp_ast.asts:
            opt_ast = ast_constraint.AstRoot([new_ast])
            new_ast = ast_constraint.AstRoot([new_ast])
            try:
                for _domain, _ast in self.__domain_optimization__(self.initial_domain, new_ast):
                    self.initial_domain, opt_ast = _domain, _ast
                self._clean_asts.add_constraint(opt_ast.asts)
            except DetectedRuntimeDependency:
                ast.add_constraint(opt_ast.asts)
            except NotEvaluateError:
                pass
            except ArgumentFunctionError:
                self._clean_asts.add_constraint(opt_ast.asts)

        if len(ast.asts) > 0:
            return self.__advance_space__(ast)
        return self

    def __build_constraint__(self, func):

        func_data = inspect.getfullargspec(func)
        args = [ast_constraint.SelfNode()]

        defaults = [] if func_data.defaults is None else func_data.defaults

        args += [ast_natural.IndexSelf(i)
                 for i in range(len(func_data.args) - 1 - len(defaults))]

        args += [ast_natural.SpaceSelfNode(item.space) for item in defaults]

        return func(*args)

    def __domain_optimization__(self, domain, ast_result):
        for visitor in reversed(self.visitor_layers):
            if not visitor.do_domain_optimization:
                continue
            # All visitors modifier the ast except the last one
            # The last one return the restricted domain
            ast_result, domain = visitor.domain_optimization(
                ast_result, domain)

            yield domain, ast_result

    def __advance_space__(self, ast):
        ss = SearchSpace(
            domain=self.initial_domain,
            distribute_like=self.__distribute_like__,
            sampler=self._distribution,
            ast=ast,
            clean_asts=self._clean_asts,
            layers=self.visitor_layers,
        )
        ss.space_name = self.space_name
        ss.set_hash(hash(self))
        return ss

    #################################################################
    #                                                               #
    #                     Sample Generate                           #
    #                                                               #
    #################################################################

    def get_sample(self, context=None, local_domain=None):
        config = SearchSpaceConfig(printer=DefaultPrinter())
        printer = config.printer_class

        printer.init_search(hash(self), self.space_name)

        if not context is None:
            cache_value = context.get_sampler_value(self)
            if not cache_value is None:
                printer.sample_value(cache_value, True)
                return cache_value, context

            precess_is_initialize = context.check_sampling_status(self)
            if not precess_is_initialize is None:
                raise CircularDependencyDetected(
                    f'in {self.__class__.__name__}')
        else:
            context = SamplerContext(self.space_name)

        context.registry_init_sampler_process(self)
        printer.context_name(context)

        domain = self.initial_domain if local_domain is None else local_domain
        printer.domain_init(domain)

        printer.tabs += 1
        domain, ast_result = self.__domain_filter__(domain, context)
        printer.tabs -= 1

        sample_index = 0
        while True:

            printer.tabs += 1
            sample, sample_context = self.__sampler__(domain, context)
            printer.tabs -= 1

            try:
                self.__check_sample__(sample, ast_result, sample_context)

                context.registry_sampler(self, sample)
                printer.sample_value(sample)
                return sample, context

            except InvalidSampler as e:
                if not config.replay_nums is None and config.replay_nums <= sample_index:
                    raise e

                config.attempts.append(sample)
                try:
                    domain = domain != sample
                except:
                    pass

                printer.sample_error(sample, e.text, sample_index, domain)

            sample_index += 1

    def __domain_filter__(self, domain, context):
        return domain, self._clean_asts

    def __sampler__(self, domain, context):
        if domain.is_invalid():
            raise InvalidSpaceConstraint(
                "the constraints change to domain in invalid")
        return domain.get_sample(self._distribution), context

    def __check_sample__(self, sample, ast_result, context):
        visitors.ValidateSampler().transform_to_check_sample(ast_result, sample, context)


class SearchSpace(BasicSearchSpace):

    def __init__(self, domain, distribute_like, sampler, ast, clean_asts, layers, _hash=None) -> None:
        super().__init__(domain, distribute_like, sampler)
        self.ast_constraint: ast_constraint.AstRoot = ast
        self._clean_asts: ast_constraint.AstRoot = clean_asts
        self.visitor_layers = layers

    def __ast_init_filter__(self):
        return self.ast_constraint

    def __ast_result_filter__(self, result):
        return result + self._clean_asts

    def __domain_filter__(self, domain, context):
        ast_result = self.__ast_init_filter__()
        if len(ast_result.asts) == 0:
            return domain, self.__ast_result_filter__(ast_result)

        config = SearchSpaceConfig(printer=DefaultPrinter())
        printer = config.printer_class

        domain = copy(domain)

        printer.ast_transformation(domain, ast_result)

        for visitor in reversed(self.visitor_layers):
            if not visitor.do_transform_to_modifier:
                continue
            # All visitors modifier the ast except the last one
            # The last one return the restricted domain

            ast_result, domain = visitor.transform_to_modifier(
                ast_result, domain, context)

            printer.ast_transformation(
                domain, ast_result, visitor.__class__.__name__)

        return domain, self.__ast_result_filter__(ast_result)

    def __advance_space__(self, ast: ast_constraint.AstRoot):
        self.ast_constraint.asts += ast.asts
        return self

    def __copy__(self):
        return type(self)(
            self.initial_domain,
            distribute_like=self.__distribute_like__,
            sampler=None,
            ast=ast_constraint.AstRoot(copy(self.ast_constraint.asts)),
            clean_asts=ast_constraint.AstRoot(copy(self._clean_asts.asts)),
            layers=copy(self.visitor_layers)
        )



from typing import Generic, Tuple, Type, TypeVar, List
from search_space.spaces.build_in_spaces import ListSearchSpace, SpacesManager, BasicCategoricalSearchSpace, TensorSearchSpace
from search_space.spaces import SearchSpace
from search_space.context_manager import SamplerContext
T = TypeVar("T")


class TypeBuilder(Generic[T]):

    def __init__(self, space: SearchSpace) -> None:
        super().__init__()
        self._space = space
        self.is_tensor_type = False

    @property
    def space(self):
        return self._space if not self.is_tensor_type else self.__space

    def __call__(self, constraints=None, min=None, max=None, options=None, distribute_like=None):

        kwarg = {}

        if not distribute_like is None:
            kwarg['distribute_like'] = distribute_like

        if not options is None:
            def f(x): return True
            if not min is None:
                def f(x): return f(x) and x > min
            if not max is None:
                def f(x): return f(x) and x < max

            options = [option for option in options if f(options)]
            result = BasicCategoricalSearchSpace(
                *options, **kwarg)
        else:

            if not min is None:
                kwarg['min'] = min
            if not max is None:
                kwarg['max'] = max

            result = self.space(**kwarg)

        if self.is_tensor_type:
            result = self._space.set_type(result)

        if constraints is None:
            return result

        return result.__ast_optimization__(constraints)

    def __getitem__(self, item):
        try:
            item = item.space
            item.space_name = item.space_name + '_tensor_size'
        except AttributeError:
            pass

        try:
            self._space.len_spaces.append(item)
        except AttributeError:
            self.__space = self._space
            self._space = ListSearchSpace([item])

        self.is_tensor_type = True
        return self


class RandomValueFunc(TypeBuilder, Generic[T]):

    def __init__(self, space: SearchSpace) -> None:
        super().__init__(space)

    def __call__(self, constraint_fun=None, min=None, max=None, options=None, distribute_like=None, context: SamplerContext = None) -> T:
        space = super().__call__(constraint_fun, min, max, options, distribute_like)
        return space.get_sample(context=context)[0]

    def __getitem__(self, item) -> 'RandomValueFunc[List[T]]':
        return super().__getitem__(item)


class RandomValueClass():
    def __getitem__(self, _type: Type[T]) -> RandomValueFunc[T]:
        types_space_class = SpacesManager().get_space_by_type(_type)
        return RandomValueFunc[T](types_space_class)


RandomValue = RandomValueClass()


class SpaceDomain(Generic[T]):
    def __init__(self, space) -> None:
        self.space: SearchSpace = space

    def __or__(self, func_constraint):
        self.space = self.space.__ast_optimization__(func_constraint)
        return self

    def get_sample(self, context: SamplerContext = None) -> Tuple[T, SamplerContext]:
        return self.space.get_sample(context=context)

    @property
    def random_sample(self) -> T:
        return self


class DomainFactory(TypeBuilder, Generic[T]):
    def __call__(self, constraint_fun=None, min=None, max=None, options=None, distribute_like=None) -> SpaceDomain[T]:
        space = super().__call__(constraint_fun, min, max, options, distribute_like)
        return SpaceDomain[T](space)

    def __getitem__(self, item) -> 'DomainFactory[List[T]]':
        return super().__getitem__(item)

    def __or__(self, func_constraint):
        return self.__call__(constraint_fun=func_constraint)


class DomainClass:
    def __getitem__(self, _type: Type[T]) -> DomainFactory[T]:
        types_space_class = SpacesManager().get_space_by_type(_type)
        return DomainFactory[T](types_space_class)


Domain = DomainClass()



from .expr_ops import ExprOpsNode
from .__base__ import check_natural_value, NaturalValuesNode
from search_space.utils.ast_tools import decorated_all_methods, check_ast_precedence


@decorated_all_methods(check_ast_precedence)
@decorated_all_methods(check_natural_value)
class AtomsOp(ExprOpsNode):

    def __getattr__(self, name):
        return GetAttr(self, name)

    def __getitem__(self, index):
        return GetItem(self, index)


class GetAttr(AtomsOp):
    pass


class GetItem(AtomsOp):
    pass


class Atoms(AtomsOp):
    def __init__(self, target) -> None:
        super().__init__(target, None)


class NaturalValue(Atoms):
    pass


NaturalValuesNode.NaturalValue = NaturalValue


class SpaceSelfNode(Atoms):
    def __init__(self, space) -> None:
        super().__init__(space)

    def __hash__(self) -> int:
        return hash(self.a)



from cProfile import label
from secrets import choice
from token import EXACT_TOKEN_TYPES
from search_space.errors import CircularDependencyDetected, InvalidSpaceDefinition, NotEvaluateError
from search_space.utils.singleton import Singleton
from ..asts import constraints
from search_space.utils import visitor
from search_space.spaces.asts import constraints
from . import VisitorLayer
from .visitor_natural_ast import NaturalAstVisitor, NaturalValuesNode


class EvalAstChecked(VisitorLayer, metaclass=Singleton):
    SELF = 'self'
    NATURAL = 'natural'
    NOT_EVAL = 'not_eval'

    def __init__(self) -> None:
        super().__init__()
        self.natural_visitor = NaturalAstVisitor()

    def transform_to_modifier(self, node, domain=None, context=None):
        self.context = context
        return self.visit(node), domain

    def transform_to_check_sample(self, node, sample, context=None):
        return self.visit(node)

    def domain_optimization(self, node, domain):
        self.context = None
        return self.visit(node), domain

    def choice_result(self, a, b, two_self_func):
        if a == self.SELF and b == self.SELF:
            return two_self_func()
        if a == self.NOT_EVAL or b == self.NOT_EVAL:
            raise NotEvaluateError()
        # if not self.SELF in [a, b]:
        #     raise NotEvaluateError()

        return a if b == self.NATURAL else b

    @visitor.on("node")
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node):
        result = constraints.AstRoot([])

        for n in node.asts:
            try:
                label, n = self.visit(n)
                if label == self.NATURAL:
                    continue

                result.add_constraint(n)
            except NotEvaluateError:
                pass

        return result

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.UniversalVariableBinaryOperation)
    def visit(self, node: constraints.UniversalVariableBinaryOperation):
        a, node_A = self.visit(node.target)
        b, node_B = self.visit(node.other)

        if isinstance(node, constraints.SegmentationExprNode):
            c, node_C = self.visit(node.value)

            s = list(set([a, b, c]))

            if (len(s) == 3):
                raise NotEvaluateError()

            def f():
                raise InvalidSpaceDefinition('Auto Constraint Definition')

            if len(s) == 1:
                if s[0] == self.SELF:
                    raise InvalidSpaceDefinition('Auto Constraint Definition')
                else:
                    raise NotEvaluateError()

            label = self.choice_result(s[0], s[1], f)
            n = type.__call__(node.__class__, node_A, node_B, node_C)
            return label, n

        def f():
            raise InvalidSpaceDefinition('Auto Constraint Definition')

        label = self.choice_result(a, b, f)
        result = type.__call__(node.__class__, node_A, node_B)
        # if a != self.SELF and b == self.SELF:
        #     result = node.inverted_op(node_A, node_B)
        # else:

        return label, result

    @visitor.when(constraints.EqualOp)
    def visit(self, node):
        a, node_A = self.visit(node.target)
        b, node_B = self.visit(node.other)

        def f():
            raise NotEvaluateError()

        return self.choice_result(a, b, f), constraints.EqualOp(node_A, node_B)

    @visitor.when(constraints.AndOp)
    def visit(self, node):
        a, node_A = self.visit(node.target)
        b, node_B = self.visit(node.other)

        def f():
            return self.SELF

        return self.choice_result(a, b, f), constraints.AndOp(node_A, node_B)

    @visitor.when(constraints.OrOp)
    def visit(self, node):
        a, node_A = self.visit(node.target)
        b, node_B = self.visit(node.other)

        def f():
            return self.SELF

        return self.choice_result(a, b, f), constraints.OrOp(node_A, node_B)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.GetItem)
    def visit(self, node):
        label, item = self.visit(node.target)
        return label, constraints.GetItem(item, node.other)

    @visitor.when(constraints.GetAttr)
    def visit(self, node):
        label, item = self.visit(node.target)
        return label, constraints.GetAttr(item, node.other)

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode):
        tag_set, node_list = set(), []
        for arg in node.args:
            label, n = self.visit(arg)
            tag_set.add(label)
            node_list.append(n)

        kws = {}
        for name, arg in node.kwargs:
            label, n = self.visit(arg)
            tag_set.add(label)
            kws[name] = n

        if (
            len(tag_set) == 3
            or self.SELF in tag_set
            or self.NOT_EVAL in tag_set
        ):
            raise NotEvaluateError()

        return self.NATURAL, constraints.FunctionNode(node.func, node_list, kws)

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.AdvancedFunctionNode):
        tag_set, node_list, self_index = set(), [], node.args_target
        for i, arg in enumerate(node.args):
            label, n = self.visit(arg)
            tag_set.add(label)
            node_list.append(n)

            if i in self_index and label != self.SELF:
                self_index.remove(i)

        kws, self_kws = {}, node.kw_target
        for name, arg in node.kwargs:
            label, n = self.visit(arg)
            tag_set.add(label)
            kws[name] = n

            if name in self_kws and label != self.SELF:
                self_kws.remove(name)

        if (self.NOT_EVAL in tag_set):
            raise NotEvaluateError()

        if not (self.SELF in tag_set):
            return self.NATURAL, constraints.FunctionNode(node.func, node_list, kws)

        return self.NATURAL, constraints.AdvancedFunctionNode(
            (self_index, self_kws),
            node.func, node_list, kws,
            node.cls
        )

    @visitor.when(constraints.SelfNode)
    def visit(self, node):
        return self.SELF, node

    @visitor.when(constraints.NaturalValue)
    def visit(self, node):

        if not self.context is None and isinstance(node.target, NaturalValuesNode):
            try:
                _ = self.natural_visitor.get_value(
                    node.target, context=self._context
                )
                return self.NATURAL, node
            except AttributeError:
                pass
            except CircularDependencyDetected:
                return self.NOT_EVAL, node

        return self.NATURAL, node

    @visitor.when(constraints.NotEvaluate)
    def visit(self, node):
        return self.NOT_EVAL, node



from .cmp_ops import CmpOpsNode
from .__base__ import check_natural_value
from search_space.utils.ast_tools import decorated_all_methods
from .segmentation_branch import SegmentationModOp
from . import __namespace__ as nsp


@decorated_all_methods(check_natural_value)
class ExprOpsNode(CmpOpsNode, metaclass=nsp.ExprNode):

    def __add__(self, other):
        return AddOp(self, other)

    def __radd__(self, other):
        return AddOp(other, self)

    def __mul__(self, other):
        return MultiOp(self, other)

    def __rmul__(self, other):
        return MultiOp(other, self)

    def __sub__(self, other):
        return SubOp(self, other)

    def __rsub__(self, other):
        return SubOp(other, self)

    def __truediv__(self, other):
        return DivOp(self, other)

    def __rtruediv__(self, other):
        return DivOp(other, self)

    def __mod__(self, other):
        return SegmentationModOp(self, other)

    def __rdivmod__(self, other):
        return SegmentationModOp(other, self)


class AddOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} + ( {type(self.other).__name__} {op} ...'

    def op(self):
        return lambda x, y: x + y


class SubOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} - ( {type(self.other).__name__} {op} ...'

    def op(self):
        return lambda x, y: x - y


class MultiOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} * ( {type(self.other).__name__} {op} ...'

    def op(self):
        return lambda x, y: x * y


class DivOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} / ( {type(self.other).__name__} {op} ...'

    def op(self):
        return lambda x, y: x / y



from xml import dom
from search_space.sampler.factory import SamplerFactory
from search_space.spaces import SearchSpace, BasicSearchSpace
from search_space.sampler.distribution_names import UNIFORM
from search_space.spaces.visitors import visitors
from search_space.spaces.build_in_spaces.space_manager import SpacesManager
from search_space.spaces.domains.categorical_domain import CategoricalDomain

__all__ = [
    "BasicCategoricalSearchSpace",
    "BasicBooleanSearchSpace",
]


@SpacesManager.registry(str)
class BasicCategoricalSearchSpace(BasicSearchSpace):
    def __init__(self, *domain, distribute_like=UNIFORM) -> None:
        super().__init__(CategoricalDomain(domain), distribute_like=distribute_like)

    # def __advance_space__(self, ast):

    #     advance_space = CategoricalSearchSpace(
    #         self.initial_domain, distribute_like=self.__distribute_like__)

    #     advance_space.ast_constraint = ast
    #     return advance_space


# class CategoricalSearchSpace(SearchSpace):
#     def __init__(self, domain, distribute_like=UNIFORM) -> None:
#         super().__init__(domain, distribute_like=distribute_like)
#         self.visitor_layers.append(visitors.DomainModifierVisitor())

#     def __sampler__(self, domain, context):
#         return domain.get_sample(self._distribution), context

#     def __domain_filter__(self, domain, context):
#         c_domain = CategoricalDomain(domain)
#         return super().__domain_filter__(c_domain, context)

#     def __copy__(self):
#         clone = CategoricalSearchSpace(*self.initial_domain)
#         clone.set_sampler(SamplerFactory().create_sampler(
#             self._distribution.__distribute_name__, search_space=clone))
#         return clone


@SpacesManager.registry(bool)
class BasicBooleanSearchSpace(BasicSearchSpace):
    def __init__(self, *domain, distribute_like=UNIFORM) -> None:
        if len(domain) == 0:
            domain = [True, False]

        super().__init__(CategoricalDomain(domain), distribute_like=distribute_like)

    # def __advance_space__(self, ast):
    #     advance_space = BooleanSearchSpace(
    #         self.initial_domain, distribute_like=self.__distribute_like__)

    #     advance_space.ast_constraint = ast
    #     return advance_space


# class BooleanSearchSpace(CategoricalSearchSpace):
#     def __init__(self, domain, distribute_like=UNIFORM) -> None:
#         super().__init__(domain, distribute_like=distribute_like)

#     def __copy__(self):
#         clone = BooleanSearchSpace()
#         clone.set_sampler(SamplerFactory().create_sampler(
#             self._distribution.__distribute_name__, search_space=clone))
#         return clone



from search_space.errors import InvalidSpaceConstraint, InvalidSpaceDefinition
from search_space.spaces.domains.categorical_domain import CategoricalDomain
from search_space.spaces.domains.__base__ import NumeralDomain
from .bached_domain import BachedDomain
from search_space.sampler import Sampler
from . import __namespace__ as nsp
from typing import Iterable
from search_space.utils.itertools import is_iterable


class ContinuosDomain(NumeralDomain, metaclass=nsp.ContinuosDomain):
    def __init__(self, _min, _max) -> None:
        self.min, self.max = _min, _max

    def __copy__(self):
        return ContinuosDomain(self.min, self.max)

    def get_sample(self, sampler: Sampler):
        return sampler.get_float(self.min, self.max)

    @property
    def limits(self):
        return (self.min, self.max)

    #################################################################
    #                                                               #
    #                     Class Operations                          #
    #                                                               #
    #################################################################

    def __getattr__(self, name):
        raise InvalidSpaceConstraint("Numeral value don't has attributes")

    #################################################################
    #                                                               #
    #                     List Operations                           #
    #                                                               #
    #################################################################

    def __getitem__(self, index):
        raise InvalidSpaceConstraint("Numeral value isn't indexed")

    #################################################################
    #                                                               #
    #                 Binary Compare Operations                     #
    #                                                               #
    #################################################################

    def __eq__(self, other):
        if is_iterable(other):
            return CategoricalDomain([item for item in other if self.min <= item and item <= self.max])

        if other > self.max or other < self.min:
            raise InvalidSpaceDefinition(
                f"{other} isn't intro the [{self.min}, {self.max}]")

        self.max = other
        self.min = other

        return self

    def __ne__(self, other):
        return self
        # if is_iterable(other):
        #     other = [item for item in other if self.min <=
        #              item and item <= self.max]
        #     other = list(set(other))

        #     if not self.min in other:
        #         other.append(self.min)

        #     if not self.max in other:
        #         other.append(self.max)

        #     other.sort()

        #     if len(other) == 2:
        #         return self

        #     return BachedDomain(
        #         * [ContinuosDomain(other[i-1], other[i]) for i in range(1, len(other))]
        #     )

        # "If other is out of current domain, it will never enter and current domain is right"
        # if other > self.max or other < self.min:
        #     return self

        # """If other is equal that one of limit, we can't limit
        #    to domain only can check this condition after we sampler"""
        # if other == self.max or self.min == other:
        #     return self

        # return BachedDomain(ContinuosDomain(self.min, other), ContinuosDomain(other, self.max))

    def __lt__(self, other):
        if is_iterable(other):
            other = min(other)

        if self.min > other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are graters that {other}")

        self.max = min(other, self.max)
        return self

    def __gt__(self, other):
        if is_iterable(other):
            other = max(other)

        if self.max < other:
            raise InvalidSpaceDefinition(
                f"All values intro [{self.min}, {self.max}] are less that {other}")

        self.min = max(other, self.min)
        return self

    def __ge__(self, other):
        return self.__gt__(other)

    def __le__(self, other):
        return self.__lt__(other)

    def __or__(self, __o):
        return BachedDomain(self, __o)

    def __mod_eq__(self, factor, value):
        return self

    def __mod_neq__(self, factor, value):
        return self

    def __mod_lt__(self, factor, value):
        return self

    def __mod_gt__(self, factor, value):
        return self

    def __mod_ge__(self, factor, value):
        return self

    def __mod_le__(self, factor, value):
        return self



from itertools import zip_longest
from typing import Iterable, Iterator


def mapper(a, b, f):
    for item in a:
        yield f(item, b)


def rmapper(a, b, f):
    try:
        for item in b:
            yield f(a, item)
    except TypeError:
        return f(a, b)


def map_reduce(a, b, f):
    try:
        iter_ = zip(a, b)

    except TypeError:
        if isinstance(a, Iterator):
            return mapper(a, b, f)
        else:
            return rmapper(a, b, f)

    for item_a, item_b in iter_:
        yield f(item_a, item_b)


def map_reduce_longest(a, b, f, zero=0):
    try:
        iter_ = zip_longest(a, b, fillvalue=zero)

    except TypeError:
        if isinstance(a, Iterator):
            return mapper(a, b, f)
        else:
            return rmapper(a, b, f)

    for item_a, item_b in iter_:
        yield f(item_a, item_b)


def compare(a, b, f):
    while True:
        try:
            item = next(a)
        except TypeError:
            item = a

        try:
            list_ = list(b)
        except TypeError:
            yield f(item, b)

        result = True
        for item_b in list_:
            result &= f(item, item_b)

        yield result


def filter(a, b, f):
    while True:
        try:
            item = next(a)
        except TypeError:
            item = a

        try:
            list_ = list(b)
        except TypeError:
            result = f(item, b)
            if not result:
                continue
            yield item

        result = True
        for item_b in list_:
            result &= f(item, item_b)

        if not result:
            continue

        yield item


def is_iterable(item):
    return isinstance(item, Iterable) and not isinstance(item, str)



class UnSupportOpError(Exception):
    def __init__(self, a, b, op: object, sub="") -> None:
        if b is None:
            super().__init__(
                f"{op} isn't support by {type(a).__name__} \n {sub}")
        else:
            super().__init__(
                f"{op} isn't support between {type(a).__name__} and {type(b).__name__} \n {sub}")


class InvalidSampler(Exception):
    @property
    def text(self):
        return self.args[0]


class InvalidSpaceDefinition(Exception):
    @property
    def text(self):
        return self.args[0]


class NotEvaluateError(Exception):
    pass


class InvalidSpaceConstraint(Exception):
    @property
    def text(self):
        return self.args[0]


class SpaceBuildingError(Exception):
    pass


class CircularDependencyDetected(Exception):
    pass


class TypeWithoutDefinedSpace(Exception):
    pass


class UndefinedSampler(Exception):
    pass


class DetectedRuntimeDependency(Exception):
    pass


class ArgumentFunctionError(Exception):
    pass


class RecursionErrorSpaceDefinition(Exception):
    pass



from .logic_ops import LogicOpsNode
from .__base__ import check_natural_value, UniversalVariableBinaryOperation
from search_space.utils.ast_tools import decorated_all_methods


class SegmentationExprNode(LogicOpsNode):
    def __init__(self, target, other, value) -> None:
        super().__init__(target, other)
        self.value = value


@decorated_all_methods(check_natural_value)
class SegmentationAtomNode(UniversalVariableBinaryOperation):
    def __and__(self, other):
        me = self == 0
        return me & other

    def __rand__(self, other):
        me = self == 0
        return other & me

    def __or__(self, other):
        me = self == 0
        return me | other

    def __or__(self, other):
        me = self == 0
        return other | me

    def __add__(self, other):
        return SegmentationAddOp(self, other)

    def __radd__(self, other):
        return SegmentationAddOp(other, self)

    def __mul__(self, other):
        return SegmentationMultiOp(self, other)

    def __rmul__(self, other):
        return SegmentationMultiOp(other, self)

    def __sub__(self, other):
        return SegmentationSubOp(self, other)

    def __rsub__(self, other):
        return SegmentationSubOp(other, self)

    def __truediv__(self, other):
        return SegmentationDivOp(self, other)

    def __rtruediv__(self, other):
        return SegmentationDivOp(other, self)

    def __mod__(self, other):
        return SegmentationModOp(self, other)

    def __divmod__(self, other):
        return SegmentationModOp(other, self)

    def __eq__(self, other):
        return SegmentationEqualOp(self.target, self.other, other)

    def __req__(self, other):
        return SegmentationEqualOp(self.target, self.other, other)

    def __ne__(self, other):
        return SegmentationNotEqualOp(self.target, self.other, other)

    def __rne__(self, other):
        return SegmentationNotEqualOp(self.target, self.other, other)

    def __lt__(self, other):
        return SegmentationLessOp(self.target, self.other, other)

    def __rlt__(self, other):
        return SegmentationGreatOp(self.target, self.other, other)

    def __gt__(self, other):
        return SegmentationGreatOp(self.target, self.other, other)

    def __rgt__(self, other):
        return SegmentationLessOp(self.target, self.other, other)

    def __ge__(self, other):
        return SegmentationGreatOrEqualOp(self.target, self.other, other)

    def __rge__(self, other):
        return SegmentationLessOrEqualOp(self.target, self.other, other)

    def __le__(self, other):
        return SegmentationLessOrEqualOp(self.target, self.other, other)

    def __rle__(self, other):
        return SegmentationGreatOrEqualOp(self.target, self.other, other)

    #################################################################
    #                                                               #
    #                 SegmentationExprNodes                         #
    #                                                               #
    #################################################################


class SegmentationEqualOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} == ( {type(self.value).__name__} {op} ...'


class SegmentationNotEqualOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} != ( {type(self.value).__name__} {op} ...'


class SegmentationGreatOrEqualOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} >= ( {type(self.value).__name__} {op} ...'


class SegmentationGreatOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} > ( {type(self.value).__name__} {op} ...'


class SegmentationLessOrEqualOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} <= ( {type(self.value).__name__} {op} ...'


class SegmentationLessOp(SegmentationExprNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % {type(self.other).__name__} < ( {type(self.value).__name__} {op} ...'

    #################################################################
    #                                                               #
    #                 SegmentationAtomNodes                         #
    #                                                               #
    #################################################################


class SegmentationAddOp(SegmentationAtomNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} + ( {type(self.other).__name__} {op} ...'


class SegmentationSubOp(SegmentationAtomNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} - ( {type(self.other).__name__} {op} ...'


class SegmentationMultiOp(SegmentationAtomNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} * ( {type(self.other).__name__} {op} ...'


class SegmentationDivOp(SegmentationAtomNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} / ( {type(self.other).__name__} {op} ...'


class SegmentationModOp(SegmentationAtomNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.target).__name__} % ( {type(self.other).__name__} {op} ...'



from unittest import result
from search_space.context_manager.runtime_manager import SearchSpacePrinter
from search_space.errors import UnSupportOpError
from search_space.utils.singleton import Singleton
from search_space.utils import visitor
from search_space.spaces.asts import constraints
from typing import Iterable
from search_space.utils.itertools import is_iterable


class Color:
    @staticmethod
    def formatter(text, color): return f'\x1b[{color}m{text}\x1b[0m'

    @staticmethod
    def f_red(text):
        return Color.formatter(text, 31)

    @staticmethod
    def f_green(text):
        return Color.formatter(text, 32)

    @staticmethod
    def f_blue(text):
        return Color.formatter(text, 96)

    @staticmethod
    def f_yellow(text):
        return Color.formatter(text, 33)

    @staticmethod
    def b_red(text):
        return Color.formatter(f'\x1b[30m{text}', 41)

    @staticmethod
    def b_green(text):
        return Color.formatter(f'\x1b[30m{text}', 42)

    @staticmethod
    def b_blue(text):
        return Color.formatter(f'\x1b[30m{text}', 44)

    @staticmethod
    def b_yellow(text):
        return Color.formatter(f'\x1b[30m{text}', 43)


class DefaultPrinter(SearchSpacePrinter, metaclass=Singleton):

    def __init__(self) -> None:
        self.tabs = 0
        self.ast_space = '|||'

    def init_search(self, id_space, name_space):
        tabs = self.tabs * '\t'

        print(f"""{tabs} _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
{tabs}| Init Sample By {Color.f_green(name_space)} Space Id: {Color.f_red(id_space)}""")

    def context_name(self, context):
        tabs = self.tabs * '\t'

        names = []
        while not context.father is None:
            names.append(context.name)
            context = context.father

        names.append(context.name)
        names.reverse()

        print(
            f"""{tabs}| {Color.b_yellow('Context')}: {' -> '.join(names)}""")

    def domain_init(self, domain):

        tabs = self.tabs * '\t'

        if isinstance(domain, type):
            name = domain.__name__
        else:
            name = domain.__class__.__name__

        try:
            limits = domain.limits
        except:
            limits = None

        print(
            f"""{tabs}| Sampler Domain Init [Domain Type: {name}] [Domain Limits: {Color.f_red(limits)}]""")

    def sample_value(self, value, caching_value=False):
        tabs = self.tabs * '\t'

        if type(value) in [list, tuple]:
            result = "[\n"
            for i, item in enumerate(value):
                result += f'{tabs}| {i} -> {item},\n'
            result += f'{tabs}| ]'
        else:
            result = value

        print(f"""{tabs}| Sample Result [Is Caching: {Color.f_blue(caching_value)}]
{tabs}| {Color.b_green('Value')}: {result}
{tabs}|_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _""")

    def sample_error(self, sample, error, sample_num, domain):

        tabs = self.tabs * '\t'

        if is_iterable(sample):
            result = "[\n"
            for i, item in enumerate(sample):
                result += f'{tabs}| {i} -> {item},\n'
            result += f'{tabs}| ]'
        else:
            result = sample

        if isinstance(domain, type):
            name = domain.__name__
        else:
            name = domain.__class__.__name__

        try:
            limits = domain.limits
        except:
            limits = None

        print(f"""{tabs}{Color.b_red('_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _')}
{tabs}| Invalid Sampler [Sampler: {result}] [Sample Numbs: {sample_num}] [Domain Type: {name}] [Domain Limits: {Color.f_red(limits)}]
{tabs}| Error: {error}
{tabs}{Color.b_red('|_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _')}""")

    def ast_transformation(self, domain, ast, visitor_name=''):
        tabs = self.tabs * '\t'

        if isinstance(domain, type):
            name = domain.__name__
        else:
            name = domain.__class__.__name__

        try:
            limits = domain.limits
        except:
            limits = None

        print(f"""{tabs}_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
{tabs}{Color.b_green(f'| AST Transform: {visitor_name}]')} [Domain Type: {name}][Domain Limits: {Color.f_red(limits)}]""")
        self.pivot_tab = self.tabs
        self.visit(ast)
        print(f"""{tabs}{Color.b_blue('| ')}_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _""")

    @visitor.on('node')
    def visit(self, node):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node: constraints.AstRoot):
        self.tabs += 1

        for n in node.asts:
            self.visit(n)

        self.tabs -= 1

    @visitor.when(constraints.UniversalVariableBinaryOperation)
    def visit(self, node):
        init_tab = self.pivot_tab * '\t'
        tabs = (self.tabs - self.pivot_tab) * self.ast_space

        print(
            f"""{init_tab}{Color.b_blue(f'| ')}{tabs} {node.__class__.__name__}:""")
        self.tabs += 1
        self.visit(node.target)
        self.visit(node.other)
        self.tabs -= 1

    @visitor.when(constraints.GetItem)
    def visit(self, node):
        init_tab = self.pivot_tab * '\t'
        tabs = (self.tabs - self.pivot_tab) * self.ast_space

        print(
            f"""{init_tab}{Color.b_blue(f'| ')}{tabs} {node.__class__.__name__}:""")
        self.tabs += 1
        self.visit(node.target)
        self.visit(node.other)
        self.tabs -= 1

    @visitor.when(constraints.GetAttr)
    def visit(self, node):
        init_tab = self.pivot_tab * '\t'
        tabs = (self.tabs - self.pivot_tab) * self.ast_space

        print(
            f"""{init_tab}{Color.b_blue(f'| ')}{tabs} {node.__class__.__name__}:""")
        self.tabs += 1
        self.visit(node.target)
        self.visit(node.other)
        self.tabs -= 1

    @visitor.when(constraints.SelfNode)
    def visit(self, node):
        init_tab = self.pivot_tab * '\t'
        tabs = (self.tabs - self.pivot_tab) * self.ast_space

        print(
            f"""{init_tab}{Color.b_blue(f'| ')}{tabs} {Color.f_blue(node.__class__.__name__)}""")

    @visitor.when(constraints.NaturalValue)
    def visit(self, node):
        init_tab = self.pivot_tab * '\t'
        tabs = (self.tabs - self.pivot_tab) * self.ast_space

        value = node.target

        try:
            value = (value.space_name, hash(value.a))
        except UnSupportOpError:
            value = type(value)
        except AttributeError:
            pass

        print(
            f"""{init_tab}{Color.b_blue(f'| ')}{tabs} {Color.f_yellow(node.__class__.__name__)}({value})""")

    # @visitor.when(FunctionNode)
    # def visit(self, node: FunctionNode, current_index):
    #     new_args = []
    #     for arg in node.args:
    #         new_args.append(self.visit(arg, current_index))

    #     new_kw = {}
    #     for name, arg in node.kwargs:
    #         new_kw[name] = self.visit(arg, current_index)

    #     return FunctionNode(node.func, new_args, new_kw)
